import abc
import copy
import itertools
import warnings

from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidExpressionError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidObjectError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLMissingCPFDefinitionError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLRepeatedVariableError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLUndefinedVariableError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLValueOutOfRangeError

from pyRDDLGym.Core.Compiler.RDDLModel import RDDLGroundedModel
from pyRDDLGym.Core.Parser.expr import Expression

AGGREG_OP_TO_STRING_DICT = {
    'prod': '*',
    'sum': '+',
    'avg': '+',
    'minimum': 'min',
    'maximum': 'max',
    'forall': '^',
    'exists': '|'
}


class Grounder(metaclass=abc.ABCMeta):
    '''Base class for all grounder classes.'''

    @abc.abstractmethod
    def Ground(self) -> RDDLGroundedModel:
        '''Produces a grounded representation of the current RDDL.'''
        pass


class RDDLGrounder(Grounder):
    '''Standard class for grounding RDDL pvariables. Does not support new 
    languages features currently.'''

    def __init__(self, RDDL_AST) -> None:
        '''Creates a new grounder object for grounding the specified RDDL file.
        '''
        super(RDDLGrounder, self).__init__()
        self.AST = RDDL_AST
        self.objects = {}
        self.objects_rev = {}
        self.pvar_to_type = {pvar.name: pvar.range 
                             for pvar in self.AST.domain.pvariables}
        self.gvar_to_type = {}
        self.gvar_to_pvar = {}
        self.nonfluents = {}
        self.states = {}
        self.statesranges = {}
        self.nextstates = {}
        self.prevstates = {}
        self.initstates = {}
        self.dynamicstate = {}
        self.actions = {}
        self.actionsranges = {}
        self.cpfs = {}
        self.cpforder = {0: []}
        self.gvar_to_cpforder = {}
        self.derived = {}
        self.interm = {}
        self.observ = {}
        self.observranges = {}
        self.reward = None
        self.terminals = []
        self.preconditions = []
        self.invariants = []
        
        self.index_of_object = {}
        self.enums = set()
        
        self.param_types = {}
        self.variable_types = {}
        self.variable_ranges = {}
        self.grounded_names = {}   
        self.default_values = {}

    def Ground(self) -> RDDLGroundedModel:
        self._extract_objects()
        self._ground_non_fluents()
        self._ground_pvariables_and_cpf()
        self._ground_init_state()
        # self._groundPvariables()
        self.reward = self._scan_expr_tree(self.AST.domain.reward, {})  # empty args dictionary
        self._ground_constraints()
        
        self._extract_variable_information()

        # update model object
        model = RDDLGroundedModel()
        model.states = self.states
        model.actions = self.actions
        model.nonfluents = self.nonfluents
        model.next_state = self.nextstates
        model.prev_state = self.prevstates
        model.init_state = self.initstate
        model.cpfs = self.cpfs  # changed to mapping name -> (<params>, <expressions>) on Dec 17 (Mike)
        model.cpforder = self.cpforder
        model.gvar_to_cpforder = self.gvar_to_cpforder
        model.reward = self.reward
        model.terminals = self.terminals
        model.preconditions = self.preconditions
        model.invariants = self.invariants
        model.derived = self.derived
        model.interm = self.interm
        model.observ = self.observ
        model.observranges = self.observranges
        model.objects = {} #self.objects
        model.objects_rev = {} #self.objects_rev
        model.actionsranges = self.actionsranges
        model.statesranges = self.statesranges
        model.gvar_to_type = self.gvar_to_type
        model.pvar_to_type = self.pvar_to_type
        model.gvar_to_pvar = self.gvar_to_pvar
        model.max_allowed_actions = self._ground_max_actions()
        model.horizon = self._ground_horizon()
        model.discount = self._ground_discount()
        
        model.SetAST(self.AST)
        
        model.param_types = self.param_types
        model.variable_types = self.variable_types
        model.variable_ranges = self.variable_ranges
        
        model.index_of_object = self.index_of_object
        model.enums = self.enums   
        model.grounded_names = self.grounded_names
        model.default_values = self.default_values
        
        return model

    def _ground_horizon(self):
        horizon = self.AST.instance.horizon
        if not (horizon >= 0):
            raise RDDLValueOutOfRangeError(
                f'Rollout horizon {horizon} in the instance is not >= 0.')
        return horizon

    def _ground_max_actions(self):
        numactions = self.AST.instance.max_nondef_actions
        if numactions == 'pos-inf':
            return len(self.actions)
        else:
            return int(numactions)

    def _ground_discount(self):
        discount = self.AST.instance.discount
        if not (0. <= discount):
            raise RDDLValueOutOfRangeError(
                f'Discount factor {discount} in the instance is not >= 0')
        return discount

    def _extract_objects(self):
        self.objects = {}
        self.objects_rev = {}
        if self.AST.non_fluents.objects \
        and self.AST.non_fluents.objects[0] is not None:
            for obj_type in self.AST.non_fluents.objects:
                self.objects[obj_type[0]] = obj_type[1]
                for obj in obj_type[1]:
                    self.objects_rev[obj] = obj_type[0]
        self.index_of_object = {}
        self.enums = set()

    def _ground_objects(self, args):
        objects_by_type = []
        for obj_type in args:
            if obj_type not in self.objects:
                raise RDDLTypeError(
                    f'Object type <{obj_type}> is not defined, '
                    f'should be one of <{set(self.objects.keys())}>.')
            objects_by_type.append(self.objects[obj_type])
        return itertools.product(*objects_by_type)
    
    def _append_variation_to_name(self, base_name, variation):
        objects = RDDLGroundedModel.OBJECT_SEP.join(variation)
        return base_name + RDDLGroundedModel.FLUENT_SEP + objects
        
    def _generate_grounded_names(self, base_name, variation_list,
                                 return_grounding_param_dict=False):
        PRIME = RDDLGroundedModel.NEXT_STATE_SYM
        all_grounded_names = []
        prime_var = PRIME in base_name
        base_name = base_name.strip(PRIME)
        grounded_name_to_params_dict = {}
        for variation in variation_list:
            grounded_name = self._append_variation_to_name(base_name, variation)
            if prime_var: 
                grounded_name += PRIME
            all_grounded_names.append(grounded_name)
            grounded_name_to_params_dict[grounded_name] = variation
        if return_grounding_param_dict:
            return all_grounded_names, grounded_name_to_params_dict
        else:  
            # in some calls to _generate_name, we do not care about the param dict
            return all_grounded_names

    def _ground_non_fluents(self):
        if not hasattr(self.AST.non_fluents, 'init_non_fluent'):
            return
        
        valid_non_fluents = {pvar.name 
                             for pvar in self.AST.domain.pvariables 
                             if pvar.is_non_fluent()}
        for init_vals in self.AST.non_fluents.init_non_fluent:
            pvar_name = init_vals[0][0]
            vtype = self.pvar_to_type[pvar_name]
            if pvar_name not in valid_non_fluents:
                warnings.warn(
                    f'Non-fluents block initializes undefined variable <{pvar_name}>.',
                    stacklevel=2)
            variations_list = [init_vals[0][1]]
            val = init_vals[1]
            if variations_list[0] is not None:
                name = self._generate_grounded_names(
                    pvar_name, variations_list, 
                    return_grounding_param_dict=False)[0]    
            else:
                name = pvar_name
            self.nonfluents[name] = val
            self.gvar_to_type[name] = vtype
            self.gvar_to_pvar[name] = pvar_name

    def _ground_interm(self, pvariable, cpf_set):
        cpf = None
        for cpfs in cpf_set:
            if cpfs.pvar[1][0] == pvariable.name:
                cpf = cpfs
        if cpf is not None:
            self.derived[pvariable.name] = pvariable.default
            self.cpfs[pvariable.name] = ([], cpf.expr)
            level = pvariable.level
            if level is None:
                level = 1
            if level in self.cpforder:
                self.cpforder[level].append(pvariable.name)
            else:
                self.cpforder[level] = [pvariable.name]

    def _ground_pvariables_and_cpf(self):
        PRIME = RDDLGroundedModel.NEXT_STATE_SYM
        all_grounded_state_cpfs = []
        all_grounded_interim_cpfs = []
        all_grounded_derived_cpfs = []
        all_grounded_observ_cpfs = []
        for pvariable in self.AST.domain.pvariables:
            name = pvariable.name
            vtype = self.pvar_to_type[name]
            if pvariable.arity > 0:
                # In line below, if we leave as an iterator object
                # will be empty after one iteration. Hence "list(.)"
                variations = list(self._ground_objects(pvariable.param_types))
                grounded, grounded_name_to_params_dict = self._generate_grounded_names(
                    name, variations, 
                    return_grounding_param_dict=True)
            else:
                grounded = [name]
                grounded_name_to_params_dict = {name: []}
            
            for gvar_name in grounded:
                self.gvar_to_pvar[gvar_name] = name
            
            # todo merge martin's code check for abuse of arity                
            if pvariable.fluent_type == 'non-fluent':
                for g in grounded:
                    if g not in self.nonfluents:
                        self.nonfluents[g] = pvariable.default
                    self.default_values[g] = pvariable.default
                    self.gvar_to_type[g] = vtype
                
            elif pvariable.fluent_type == 'action-fluent':
                for g in grounded:
                    self.actions[g] = pvariable.default
                    self.default_values[g] = pvariable.default
                    self.actionsranges[g] = pvariable.range
                    self.gvar_to_type[g] = vtype
              
            elif pvariable.fluent_type == 'state-fluent':
                cpf = None
                next_state = name + PRIME
                for cpfs in self.AST.domain.cpfs[1]:
                    if cpfs.pvar[1][0] == next_state:
                        cpf = cpfs
                        break
                if cpf is None:
                    raise RDDLMissingCPFDefinitionError(
                        f'CPF <{name}> is missing a valid definition.')
    
                for g in grounded:
                    grounded_cpf = self._ground_single_cpf(
                        cpf, g, grounded_name_to_params_dict[g])
                    all_grounded_state_cpfs.append(grounded_cpf)
                    next_state = g + PRIME  # update to grounded version, satisfied single-variables too (i.e. not a type)
                    self.states[g] = pvariable.default
                    self.default_values[g] = pvariable.default
                    self.statesranges[g] = pvariable.range
                    self.nextstates[g] = next_state
                    self.prevstates[next_state] = g
                    self.cpfs[next_state] = ([], grounded_cpf.expr)
                    self.cpforder[0].append(g)
                    self.gvar_to_cpforder[g] = 0
                    self.gvar_to_type[g] = vtype
                    self.gvar_to_pvar[next_state] = name
                    self.gvar_to_type[next_state] = vtype
                    
            elif pvariable.fluent_type == 'derived-fluent':
                cpf = None
                for cpfs in self.AST.domain.derived_cpfs:
                    if cpfs.pvar[1][0] == name:
                        cpf = cpfs
                        break
                if cpf is None:
                    raise RDDLMissingCPFDefinitionError(
                        f'CPF <{name}> is missing a valid definition.')
                    
                for g in grounded:
                    grounded_cpf = self._ground_single_cpf(
                        cpf, g, grounded_name_to_params_dict[g])
                    all_grounded_derived_cpfs.append(grounded_cpf)
                    self.derived[g] = pvariable.default
                    self.default_values[g] = pvariable.default
                    self.cpfs[g] = ([], grounded_cpf.expr)
                    level = pvariable.level
                    if level is None:
                        level = 1
                    if level in self.cpforder:
                        self.cpforder[level].append(g)
                    else:
                        self.cpforder[level] = [g]
                    self.gvar_to_cpforder[g] = level
                    self.gvar_to_type[g] = vtype
    
            elif pvariable.fluent_type == 'interm-fluent':
                cpf = None
                for cpfs in self.AST.domain.intermediate_cpfs:
                    if cpfs.pvar[1][0] == name:
                        cpf = cpfs
                        break
                if cpf is None:
                    raise RDDLMissingCPFDefinitionError(
                        f'CPF <{name}> is missing a valid definition.')
                    
                for g in grounded:
                    grounded_cpf = self._ground_single_cpf(
                        cpf, g, grounded_name_to_params_dict[g])
                    all_grounded_interim_cpfs.append(grounded_cpf)
                    self.interm[g] = pvariable.default
                    self.default_values[g] = pvariable.default
                    self.cpfs[g] = ([], grounded_cpf.expr)
                    level = pvariable.level
                    if level is None:
                        level = 1
                    if level in self.cpforder:
                        self.cpforder[level].append(g)
                    else:
                        self.cpforder[level] = [g]
                    self.gvar_to_type[g] = vtype
                    self.gvar_to_cpforder[g] = level
                    
            elif pvariable.fluent_type == 'observ-fluent':
                cpf = None
                for cpfs in self.AST.domain.observation_cpfs:
                    if cpfs.pvar[1][0] == name:
                        cpf = cpfs
                        break
                if cpf is None:
                    raise RDDLMissingCPFDefinitionError(
                        f'CPF <{name}> is missing a valid definition.')
                    
                for g in grounded:
                    grounded_cpf = self._ground_single_cpf(
                        cpf, g, grounded_name_to_params_dict[g])
                    all_grounded_observ_cpfs.append(grounded_cpf)
                    self.observ[g] = pvariable.default
                    self.default_values[g] = pvariable.default
                    self.observranges[g] = pvariable.range
                    self.cpfs[g] = ([], grounded_cpf.expr)
                    self.cpforder[0].append(g)
                    self.gvar_to_type[g] = vtype
                    self.gvar_to_cpforder[g] = 0

        # self.AST.domain.cpfs = (self.AST.domain.cpfs[0], all_grounded_state_cpfs
        #                        )  # replacing the previous lifted entries
        # self.AST.domain.derived_cpfs = all_grounded_derived_cpfs
        # self.AST.domain.intermediate_cpfs = all_grounded_interim_cpfs

    def _ground_single_cpf(self, cpf, variable, variable_args):
        """Map arguments to actual objects."""
        args = cpf.pvar[1][1]
        new_cpf = copy.deepcopy(cpf)
        if args is None:
            new_cpf.expr = self._scan_expr_tree(new_cpf.expr, {})
            return new_cpf
        if len(args) != len(variable_args):
            raise RDDLInvalidNumberOfArgumentsError(
                f'Ground instance <{variable}> is of arity {len(variable_args)}, '
                f'but expected to be of arity {len(args)} according to definition.')
        args_dic = dict(zip(args, variable_args))
            
        # Parse cpf w.r.t cpf args and variables.
        # Fix name.
        cpf_base_name = new_cpf.pvar[1][0]
        cpf_name_grounding_variation = [[args_dic[arg] for arg in new_cpf.pvar[1][1]]]  # must be a nested list for func call
        new_name = self._generate_grounded_names(
            cpf_base_name, cpf_name_grounding_variation)
        new_pvar = ('pvar_expr', (new_name, None))
        new_cpf.pvar = new_pvar
        new_cpf.expr = self._scan_expr_tree(new_cpf.expr, args_dic)
        return new_cpf

    def do_aggregate_expression_grounding(self, original_dict, new_variables_list,
                                          instances_list, operation_string,
                                          expression):
        """Args:
                original_dict:
                new_variables_list:
                instances_list:
                operation_string:
                expression:
            Returns: Grounded expression for aggregation operations (min, max, sum, prod, forall, exists)
        """

        new_children = []
        for instance in instances_list:
            updated_dict = copy.deepcopy(original_dict)
            updated_dict.update(zip(new_variables_list, instance))
            new_children.append(self._scan_expr_tree(expression, updated_dict))
        new_expr = Expression((operation_string, tuple(new_children)))
        return new_expr

    def _scan_expr_tree_pvar(self, expr: Expression, dic) -> Expression:
        """Ground out a pvar expression."""
        if expr.args[1] is None:
            # This is a constant: should really be etype = constant in parsed tree.
            pass
        
        elif expr.args[1]:
            variation_list = []
            for arg in expr.args[1]:
                if arg not in dic:
                    raise RDDLUndefinedVariableError(
                        f'Parameter <{arg}> is not defined in call to <{expr.args[0]}>.')
                variation_list.append(dic[arg])
            variation_list = [variation_list]
            new_name = self._generate_grounded_names(expr.args[0], variation_list)[0]
            expr = Expression(('pvar_expr', (new_name, None)))
            
        else:
            raise RDDLInvalidExpressionError(f'Malformed expression <{expr}>.')
        
        return expr

    def _scan_expr_tree_abr(self, expr, dic):
        new_children = []
        for child in expr.args:
            new_children.append(self._scan_expr_tree(child, dic))
        return Expression((expr.etype[1], tuple(new_children)))

    def _scan_expr_tree_control(self, expr, dic):
        children_list = [
            self._scan_expr_tree(expr.args[0], dic),
            self._scan_expr_tree(expr.args[1], dic),
            self._scan_expr_tree(expr.args[2], dic)
        ]
        # TODO: add default case when no "else". 
        # For now, we are safe, as else is expected in rddl
        return Expression(('if', tuple(children_list)))

    def _scan_expr_tree_func(self, expr, dic):
        new_children = []
        for child in expr.args:
            new_children.append(self._scan_expr_tree(child, dic))
        return Expression((expr.etype[0], (expr.etype[1], new_children)))  # Only one arg for abs.

    def _scan_expr_tree_aggregation(self, expr, dic):
        """Ground out an aggregation expression."""
        # TODO: as of now the code assumes all the leaf variables/constants are
        # of the right types, or can be reasonably cast into the right type
        # (eg: bool->int or v.v.).
        # However, some type checking would be nice in subsequent versions,
        # and give feedback to the language writer for debugging.
        aggreg_type = expr.etype[1]
        if aggreg_type in AGGREG_OP_TO_STRING_DICT:
            
            # Determine what the recursive op is for the aggreg type. Eg: sum = "+".
            aggreg_recursive_operation_string = AGGREG_OP_TO_STRING_DICT[aggreg_type]
            
            # TODO: only for average operation, we need to first "/ n " for all others
            # we need to decide the recursive operation symbol "+" or "*" and iterate.
            # First let's collect the instances like (?x,?y) = (x1, y3) that satisfy,
            # the set definition passed in.
            object_instances_list = []
            instances_def_args = []
            for arg in expr.args:
                if arg[0] == 'typed_var':
                    instances_def_args += list(arg[1])
            if instances_def_args:  
                
                # Then we iterate over the objects specified.
                # All even indexes (incl 0) are variable names,
                # all odd indexes are object types.
                var_key_strings_list = [
                    instances_def_args[2 * x]
                    for x in range(int(len(instances_def_args) / 2))
                ]  # Like ?x.
                object_type_list = [
                    instances_def_args[2 * x + 1]
                    for x in range(int(len(instances_def_args) / 2))
                ]
                instance_tuples = [
                    tuple([x]) for x in self.objects[object_type_list[0]]
                ]
                for var_idx in range(1, len(var_key_strings_list)):
                    instance_tuples = [
                        tuple(list(instance) + [objects])
                        for instance in instance_tuples
                        for objects in self.objects[object_type_list[var_idx]]
                    ]
                object_instances_list = instance_tuples
                expr = self.do_aggregate_expression_grounding(
                    dic, var_key_strings_list, object_instances_list,
                    aggreg_recursive_operation_string, arg
                )  # Last arg is the expression with which the aggregation is done.
                
                if aggreg_type == 'avg':
                    num_instances = len(instance_tuples)  # Needed if this is an "Avg" operation.
                    # Then the 'expr' becomes lhs argument and
                    # we add a "\ |set_size|" operation.
                    children_list = [expr, Expression(('number', num_instances))]
                    # Note "expr" would have been an aggregate sum already,
                    # the "aggreg_recursive_operation_string" is set for that.
                    expr = Expression(('/', tuple(children_list)))
            return expr

    def _scan_expr_tree(self, expr: Expression, dic) -> Expression:
        """Main dispatch method for recursively grounding the expression tree."""
        scan_expr_tree_noop = lambda expr, _: expr
        dispatch_dict = {
            'noop': scan_expr_tree_noop,
            'pvar': self._scan_expr_tree_pvar,
            'constant': scan_expr_tree_noop,
            'arithmetic': self._scan_expr_tree_abr,
            'boolean': self._scan_expr_tree_abr,
            'relational': self._scan_expr_tree_abr,
            'aggregation': self._scan_expr_tree_aggregation,
            'control': self._scan_expr_tree_control,
            # Random vars can be ground in the same way as functions.
            'func': self._scan_expr_tree_func,
            'randomvar': self._scan_expr_tree_func
        }
        if isinstance(expr, tuple):
            expression_type = 'noop'
        else:
            expression_type = expr.etype[0]
        if expression_type in dispatch_dict.keys():
            return dispatch_dict[expression_type](expr, dic)
        else:
            new_children = []
            for child in expr.args:
                new_children.append(self._scan_expr_tree(child, dic))
            # If we reached here the expression is either a +,*, or comparator (>,<),
            # or aggregator (sum, product).
            return Expression((expr.etype[1], tuple(new_children)))

    def _ground_constraints(self) -> None:
        if hasattr(self.AST.domain, 'terminals'):
            for terminal in self.AST.domain.terminals:
                self.terminals.append(self._scan_expr_tree(terminal, {}))

        if hasattr(self.AST.domain, 'preconds'):
            for precond in self.AST.domain.preconds:
                self.preconditions.append(self._scan_expr_tree(precond, {}))
    
        if hasattr(self.AST.domain, 'constraints'):
            if self.AST.domain.constraints:
                warnings.warn(
                    f'State-action constraints are not implemented '
                    f'in this RDDL version and will be ignored.',
                    stacklevel=2)
    
        if hasattr(self.AST.domain, 'invariants'):
            for inv in self.AST.domain.invariants:
                self.invariants.append(self._scan_expr_tree(inv, {}))

    def _ground_init_state(self) -> None:
        self.initstate = self.states.copy()
        if hasattr(self.AST.instance, 'init_state'):
            for init_vals in self.AST.instance.init_state:
                (key, subs), val = init_vals
                if subs:
                    key = self._append_variation_to_name(key, subs)
                if key in self.initstate:
                    self.initstate[key] = val
                else:
                    warnings.warn(
                        f'Init-state block initializes undefined state-fluent <{key}>.',
                        stacklevel=2)
                    
    # added on Dec 17 (Mike)
    def _extract_variable_information(self):
        PRIME = RDDLGroundedModel.NEXT_STATE_SYM
        var_params, var_types, var_ranges = {}, {}, {}
        for pvar in self.AST.domain.pvariables:
            
            # make sure name does not contain separators
            SEPARATORS = [RDDLGroundedModel.FLUENT_SEP, RDDLGroundedModel.OBJECT_SEP] 
            for separator in SEPARATORS:
                if separator in pvar.name:
                    raise RDDLInvalidObjectError(
                        f'Variable name <{pvar.name}> contains the '
                        f'illegal separator {separator}.')
                    
            objects = pvar.param_types
            if objects is None:
                objects = []
            if objects:
                variations = list(self._ground_objects(objects))
                grounded = self._generate_grounded_names(pvar.name, variations)
            else:
                grounded = [pvar.name]
            for name in grounded:
                if name in var_params:
                    raise RDDLRepeatedVariableError(
                        f'{pvar.fluent_type} <{pvar.name}> has the same name as '
                        f'another pvariable {var_types[name]}.')
                primed_name = (name + PRIME) if pvar.is_state_fluent() else name
                var_params[name] = var_params[primed_name] = []        
                var_types[name] = pvar.fluent_type
                if pvar.is_state_fluent():
                    var_types[primed_name] = 'next-state-fluent'                
                var_ranges[name] = var_ranges[primed_name] = pvar.range 
        grounded_names = {name: [name] for name in var_params}
          
        self.param_types, self.variable_types, self.variable_ranges = \
            var_params, var_types, var_ranges
        self.grounded_names = grounded_names
        



from pyRDDLGym.Core.Env import RDDLEnv as RDDLEnv
from pyRDDLGym.Policies.Agents import RandomAgent

from pyRDDLGym.Visualizer.WildfireViz import WilfireVisualizer

# FOLDER = 'Competition/Power_gen/'
# FOLDER = 'Competition/Mars_rover/'
# FOLDER = 'Competition/UAVs_mix/'
FOLDER = 'Competition/Wildfire/'


def main():
    steps = 100
    myEnv = RDDLEnv.RDDLEnv(domain=FOLDER + 'domain.rddl', instance=FOLDER + 'instance0.rddl', is_grounded=False)
    agent = RandomAgent(action_space=myEnv.action_space, num_actions=myEnv.NumConcurrentActions)
    # myEnv.set_visualizer(PowerGenVisualizer)
    # myEnv.set_visualizer(MarsRoverVisualizer)
    # myEnv.set_visualizer(UAVsVisualizer)
    myEnv.set_visualizer(WilfireVisualizer)

    # pprint(vars(myEnv.model))
    
    total_reward = 0
    state = myEnv.reset()
    
    for step in range(steps):

        img = myEnv.render()

        # img.save('./img_folder/drone/'+str(step)+'.png')

        action = agent.sample_action()

        next_state, reward, done, info = myEnv.step(action)
        total_reward += reward
        # print("step {}: reward: {}".format(step, reward))
        # print("state_i:", state, "-> state_f:", next_state)

        print("\nstep = {}".format(step))
        print('reward = {}'.format(reward))
        print('state = {}'.format(state))
        print('action = {}'.format(action))
        print('next_state = {}'.format(next_state))
        print('derived = {}'.format(myEnv.model.derived))
        print('interm = {}'.format(myEnv.model.interm))

        state = next_state       


    print("episode ended with reward {}".format(total_reward))


if __name__ == "__main__":
    main()



from pyRDDLGym.Core.Env import RDDLEnv as RDDLEnv
from pyRDDLGym.Core.Policies.Agents import RandomAgent

FOLDER = './Competition/Drone/Drone_con/'


def main():
    steps = 1000
    myEnv = RDDLEnv.RDDLEnv(domain=FOLDER + 'domain.rddl',
                            instance=FOLDER + 'insta0.rddl')
    agent = RandomAgent(action_space=myEnv.action_space,
                        num_actions=myEnv.numConcurrentActions)

    from pprint import pprint
    pprint(vars(myEnv.model))

    total_reward = 0
    state = myEnv.reset()
    for step in range(steps):
        myEnv.render()
        action = agent.sample_action()

        next_state, reward, done, info = myEnv.step(action)
        total_reward += reward
        # print("step {}: reward: {}".format(step, reward))
        # print("state_i:", state, "-> state_f:", next_state)
        print("\nstep = {}".format(step))
        print('reward = {}'.format(reward))
        print('state = {}'.format(state))
        print('action = {}'.format(action))
        print('next_state = {}'.format(next_state))
        print('derived = {}'.format(myEnv.model.derived))
        print('interm = {}'.format(myEnv.model.interm))
        state = next_state
    print("episode ended with reward {}".format(total_reward))


if __name__ == "__main__":
    main()



import pygraphviz as pgv
from pathlib import Path
from typing import Optional, Tuple, Dict, Union, List, Set

from pyRDDLGym.Core.Grounder.RDDLGrounder import RDDLGrounder
from pyRDDLGym.XADD.RDDLModelXADD import RDDLModelWXADD
from pyRDDLGym.Core.Parser.RDDLReader import RDDLReader
from pyRDDLGym.Core.Parser.parser import RDDLParser
from pyRDDLGym.Examples.ExampleManager import ExampleManager

COLOR = dict(
    action='olivedrab1',
    interm='sandybrown',
    derived='orchid',
    observ='orangered',
    state='lightblue',
    next_state='gold1',
    reward='firebrick1'
)

SHAPE = dict(
    action='box',
    interm='ellipse',
    derived='octagon',
    observ='ellipse',
    state='ellipse',
    next_state='ellipse',
    reward='diamond'
)

STYLE = dict(
    action='filled',
    interm='filled',
    derived='filled',
    observ='filled',
    state='filled',
    next_state='filled',
    reward='filled'
)


class Graph(pgv.AGraph):

    def __init__(self,
                 thing=None,
                 filename=None,
                 data=None,
                 string=None,
                 handle=None,
                 name="",
                 strict=True,
                 directed=False,
                 **attr):
        super().__init__(thing, filename, data, string, handle, name, 
                         strict, directed, **attr)
        self._supress_rank = False
        self._same_rank: List[List[str]] = []
    
    def set_suppress_rank(self, suppress_rank: bool):
        self._suppress_rank = suppress_rank
    
    def add_same_rank(self, nodes: List[str]):
        self._same_rank.append(nodes)
    
    def configure_graph_attributes(self, **attrs):
        self.graph_attr.update(attrs)

    def configure_node_attributes(self, **attrs):
        self.node_attr.update(attrs)

    def configure_edge_attributes(self, **attrs):
        self.edge_attr.update(attrs)

    def set_ranks(self):
        if not self._supress_rank:
            for r in self._same_rank:
                self.add_subgraph(r, rank='same')


class RDDL2Graph:

    def __init__(self, domain: str='Wildfire',
                 instance: int=0,
                 directed: bool=True,
                 strict_grouping: bool=False,
                 simulation: bool=False):

        # Read the domain and instance files
        self._domain, self._instance = domain, str(instance)
        env_info = ExampleManager.GetEnvInfo(domain)
        domain = env_info.get_domain()
        instance = env_info.get_instance(instance)
        
        # Read and parse domain and instance
        reader = RDDLReader(domain, instance)
        domain = reader.rddltxt
        parser = RDDLParser(None, False)
        parser.build()

        # Parse RDDL file
        rddl_ast = parser.parse(domain)

        # Ground domain
        grounder = RDDLGrounder(rddl_ast)
        model = grounder.Ground()

        # XADD compilation
        self.model = RDDLModelWXADD(model, simulation=simulation)
        self.model.compile()
        
        self.cpfs: Dict[str, Union[int, List[int]]] = self.model.cpfs
        self.cpfs.update({'reward': self.model.reward})
        self.cpfs.update({'terminals': self.model.terminals})
        self._directed = directed
        self._strict_grouping = strict_grouping
        self._gvar_to_node_name = {}
        self._node_name_to_gvar = {}
        self._action: Set[str] = set()
        self._state = set()
        self._next_state = set()
        self._observ = set()
        self._interm = set()
        self._derived = set()
        self._parents = {}
    
    def configure_graph(self, graph: Graph):
        graph.configure_graph_attributes(
            fontname="Helvetica",
            fontsize="16",
            ratio="auto",
            size="7.5,10",
            rankdir="LR",
            ranksep="2.00"
        )
    
    def configure_nodes(self, graph: Graph):
        graph.configure_node_attributes(
            fontsize="16",
            color='black',
        )
        for node in graph.nodes():
            color, style, shape = [''] * 3
            if node in self._state:
                color, shape, style = COLOR['state'], SHAPE['state'], STYLE['state']
            elif node in self._next_state:
                color, shape, style = COLOR['next_state'], SHAPE['next_state'], STYLE['next_state']
            elif node in self._action:
                color, shape, style = COLOR['action'], SHAPE['action'], STYLE['action']
            elif node in self._interm:
                color, shape, style = COLOR['interm'], SHAPE['interm'], STYLE['interm']
            elif node in self._observ:
                color, shape, style = COLOR['observ'], SHAPE['observ'], STYLE['observ']
            elif node in self._derived:
                color, shape, style = COLOR['derived'], SHAPE['derived'], STYLE['derived']
            elif node == 'Reward Function':
                color, shape, style = COLOR['reward'], SHAPE['reward'], STYLE['reward']
            if color or style or shape:
                node.attr['color'] = color
                node.attr['style'] = style
                node.attr['shape'] = shape
        
    def configure_edges(self, graph: Graph):
        if not self._directed:
            graph.configure_edge_attributes(fontsize="16")

    def save_dbn(self, file_name: str='example',
                 fluent: Optional[str]=None,
                 gfluent: Optional[str]=None,
                 file_type: str='pdf'):
        self.clear_cache()
        f_dir = Path(f"tmp/{self._domain}")
        f_dir.mkdir(exist_ok=True, parents=True)
        f_path = f_dir / \
            f"{file_name}{('_' + fluent) if fluent else ''}{('_' + gfluent) if gfluent else ''}_inst_{self._instance}"
        if fluent is None and gfluent is None:
            graph = self.get_graph_of_instance()
        else:
            graph = self.get_graph_of_cpf(fluent, gfluent)
        self.configure_graph(graph)
        self.configure_nodes(graph)
        self.configure_edges(graph)
        graph.set_ranks()
        graph.draw(f"{str(f_path)}.{file_type}", prog='dot')

        # Output a file
        with open(f_dir / f"{file_name}_inst_{self._instance}.txt", "w") as f:
            txt = self.get_text_repr()
            f.write(txt)
    
    def get_text_repr(self):
        """Returns a string that summarizes the DBN dependency structure of a given RDDL instance"""
        res = f"DBN dependency analysis\n\n" \
              f"Domain: {self._domain}\n" \
              f"Instance: {self._instance}\n\n"
        
        res += f"{'Grounded fluent':40}{'Parent variables':100}\n\n"
        # Ground fluents (including terminals and reward)
        for cpf, node_id in self.model.cpfs.items():
            if not node_id or cpf == 'terminals':
                continue
            parents = self.model.collect_vars(node_id)
            parents = sorted([self.get_node_name(p) for p in parents])
            node_name = self.get_node_name(cpf) if cpf != 'reward' else cpf
            cpf_str = f"{node_name:40}{', '.join(parents):100}".strip(',')
            res += f"{cpf_str}\n"
        return res

    def get_graph_of_cpf(self, fluent: str, gfluent: Optional[str]=None) -> Graph:
        assert fluent in self.model.pvar_to_type or fluent.lower() == 'reward', \
            f"Fluent {fluent} not recognized"
        assert not gfluent or (gfluent and f"{fluent}_{gfluent}" in self.model.gvar_to_type), \
            f"Grounded fluent provided but cannot be resolved"

        graph = Graph(directed=self._directed)
        if not self._strict_grouping:
            graph.set_suppress_rank(True)
        
        if fluent.lower() == 'reward':
            gvars = ['reward']
        else:
            gvars = {
                gvar for gvar, pvar in self.model.gvar_to_pvar.items() 
                    if pvar == fluent and 
                        ((not gfluent) or (gfluent and f"{fluent}_{gfluent}" == gvar))
            }
        
        for gvar in gvars:
            if gvar in self.model.interm:
                self.add_single_interm_fluent_to_graph(graph, gvar)
            elif gvar in self.model.derived:
                self.add_single_derived_fluent_to_graph(graph, gvar)
            elif gvar in self.model.states:
                self.add_single_state_fluent_to_graph(graph, gvar)
            elif gvar in self.model.observ:
                self.add_single_observ_fluent_to_graph(graph, gvar)
            elif gvar in self.model.actions:
                raise NotImplementedError("For action variables, generate the entire DBN")
            elif gvar == 'reward':
                self.add_reward_to_graph(graph)

        # Handle action nodes
        action_nodes = list(self._action)
        for act in action_nodes:
            graph.add_node(act)
        
        # Handle interm & derived fluents
        interm_derived_nodes = list(self._interm) + list(self._derived)
        interm_derived_nodes = [f"Intermediate"] + interm_derived_nodes
        graph.add_same_rank(interm_derived_nodes)
        
        # Handle observations
        obs_nodes = ["Observations"] + list(self._observ)
        if len(obs_nodes) > 1:
            graph.add_same_rank(obs_nodes)
        
        state_nodes = list(self._state)
        action_states = ["Current State and Actions"] + action_nodes + state_nodes
        graph.add_same_rank(action_states)

        # Handle reward node
        if fluent.lower() != 'reward':
            self.add_reward_to_graph(graph, fluent=fluent, gfluent=gfluent)
        
        # Put next state and reward nodes at same rank
        next_state_nodes = list(self._next_state)
        rew_ns_nodes = ['Reward Function', 'Next State and Reward'] + next_state_nodes
        graph.add_same_rank(rew_ns_nodes)

        # Setup some stratification nodes and links for strict levels
        if self._strict_grouping:
            prev_level = "Current State and Actions"
            graph.add_node(prev_level, color='white', shape='plaintext', style='bold')

            # interm & derived
            level_node = f"Intermediate"
            graph.add_edge(prev_level, level_node, color='black', style='invis', label='')
            graph.add_node(level_node, color='white', shape='plaintext', style='bold')
            prev_level = level_node
            
            node_name = "Next State and Reward"
            graph.add_node(node_name, color='white', shape='plaintext', style='bold')
            graph.add_edge(prev_level, node_name, color='black', style='invis', label='')

            # Observations (at the right-most column)
            if len(obs_nodes) > 1:  # will always contain default "Observations" node
                graph.add_node("Observations", color='white', shape='plaintext', style='bold')
                graph.add_edge(prev_level, "Observations", color='black', style='invis', label='')
                prev_level = "Observations"
        
        return graph
        
    def get_graph_of_instance(self) -> Graph:
        graph = Graph(directed=self._directed)
        if not self._strict_grouping:
            graph.set_suppress_rank(True)
        
        # Go through all actions
        for act in self.model.actions:
            node_name = self.get_node_name(act)
            self.add_node(node_name)
            graph.add_node(node_name)
        
        # Handle intermediate & derived variables
        self.add_interm_fluents_to_graph(graph)
        self.add_derived_fluents_to_graph(graph)
        interm_derived_nodes = list(self._interm) + list(self._derived)
        interm_derived_nodes = [f"Intermediate"] + interm_derived_nodes
        graph.add_same_rank(interm_derived_nodes)
        
        # Handle observations
        self.add_observ_fluents_to_graph(graph)
        obs_nodes = ["Observations"] + list(self._observ)
        if len(obs_nodes) > 1:
            graph.add_same_rank(obs_nodes)
        
        # Handle current and next state CPFs
        self.add_state_fluents_to_graph(graph)
        
        # Put all current state and action nodes at same rank
        action_nodes = list(self._action)
        state_nodes = list(self._state)
        action_states = ["Current State and Actions"] + action_nodes + state_nodes
        graph.add_same_rank(action_states)

        # Handle reward node
        self.add_reward_to_graph(graph)

        # Put Next State and Reward nodes at same rank
        next_state_nodes = list(self._next_state)
        rew_ns_nodes = ['Reward Function', 'Next State and Reward'] + next_state_nodes
        graph.add_same_rank(rew_ns_nodes)

        # Setup some stratification nodes and links if strict levels
        if self._strict_grouping:
            prev_level = "Current State and Actions"
            graph.add_node(prev_level, color='white', shape='plaintext', style='bold')

            level_node = f"Intermediate"
            graph.add_edge(prev_level, level_node, color='black', style='invis', label='')
            graph.add_node(level_node, color='white', shape='plaintext', style='bold')
            prev_level = level_node
            
            node_name = "Next State and Reward"
            graph.add_node(node_name, color='white', shape='plaintext', style='bold')
            graph.add_edge(prev_level, node_name, color='black', style='invis', label='')

            # Observations
            if len(obs_nodes) > 1:  # will always contain default "Observations" node
                graph.add_node("Observations", color='white', shape='plaintext', style='bold')
                graph.add_edge(prev_level, "Observations", color='black', style='invis', label='')
                prev_level = "Observations"
        
        return graph
            
    def add_reward_to_graph(
            self, graph: Graph, fluent: Optional[str]=None, gfluent: Optional[str]=None
    ):
        graph.add_node("Reward Function")
        parents = self.model.vars_in_rew
        for p in parents:
            pvar = self.model.gvar_to_pvar[p]
            if not fluent or \
                (fluent and pvar == fluent and not gfluent) or \
                    (gfluent and (p == f"{fluent}_{gfluent}" or p == f"{fluent}_{gfluent}'")):
                p_node_name = self.get_node_name(p)
                self.add_node(p_node_name)
                graph.add_edge(p_node_name, 'Reward Function')

    def add_single_state_fluent_to_graph(self, graph: Graph, state: str):        
        ns = self.model.next_state[state]  # Primed state variable
        
        # Get node names and add to the graph
        s_node_name = self.get_node_name(state)
        ns_node_name = self.get_node_name(ns, primed=True)
        self.add_node(s_node_name); self.add_node(ns_node_name)
        graph.add_node(s_node_name); graph.add_node(ns_node_name)
        
        # Get parents and add links
        parents = self.model.collect_vars(self.cpfs[ns])
        for p in parents:
            p_node_name = self.get_node_name(
                p, primed=p in self.model.next_state.values())
            self.add_node(p_node_name)
            graph.add_edge(
                p_node_name,
                ns_node_name,
                color='black',
                style='solid',
                label=''
            )
    
    def add_state_fluents_to_graph(self, graph: Graph):
        for st in self.model.states:
            self.add_single_state_fluent_to_graph(graph, st)

    def add_single_observ_fluent_to_graph(self, graph: Graph, observ: str):
        node_name = self.get_node_name(observ)
        self._gvar_to_node_name[observ] = node_name
        graph.add_node(node_name)
        self.add_node(node_name)

        # Get parent nodes and add links
        parents = self.model.collect_vars(self.cpfs[observ])
        for p in parents:
            p_node_name = self.get_node_name(p)
            self.add_node(p_node_name)
            graph.add_edge(p_node_name, node_name)
    
    def add_observ_fluents_to_graph(self, graph: Graph):
        for observ in self.model.observ:
            self.add_single_observ_fluent_to_graph(graph, observ)
    
    def add_single_interm_fluent_to_graph(self, graph: Graph, interm: str):
        node_name = self.get_node_name(interm)
        self._gvar_to_node_name[interm] = node_name
        self.add_node(node_name)
        graph.add_node(node_name)

        # Get parent nodes and add links
        parents = self.model.collect_vars(self.cpfs[interm])
        for p in parents:
            p_node_name = self.get_node_name(p)
            self.add_node(p_node_name)
            graph.add_edge(p_node_name, node_name)

    def add_interm_fluents_to_graph(self, graph: Graph):
        for interm in self.model.interm:
            self.add_single_interm_fluent_to_graph(graph, interm)
    
    def add_single_derived_fluent_to_graph(self, graph: Graph, derived: str):
        node_name = self.get_node_name(derived)
        self._gvar_to_node_name[derived] = node_name
        self.add_node(node_name)
        graph.add_node(node_name)

        # Get parent nodes and add links
        parents = self.model.collect_vars(self.cpfs[derived])
        for p in parents:
            p_node_name = self.get_node_name(p)
            self.add_node(p_node_name)
            graph.add_edge(p_node_name, node_name)

    def add_derived_fluents_to_graph(self, graph: Graph):
        for derived in self.model.derived:
            self.add_single_derived_fluent_to_graph(graph, derived)
    
    @staticmethod
    def get_objects(gvar: str, pvar: str) -> List[str]:
        return list(map(lambda x: '?' + x, 
                        [v_str for v_str in gvar.split(pvar)[1].split('_')[1:] if v_str]))
    
    def add_node(self, node_name: str):
        gvar = self._node_name_to_gvar[node_name]
        if gvar in self.model.prev_state.values():
            self._state.add(node_name)
        elif gvar in self.model.next_state.values():
            self._next_state.add(node_name)
        elif gvar in self.model.interm:
            self._interm.add(node_name)
        elif gvar in self.model.derived:
            self._derived.add(node_name)
        elif gvar in self.model.observ:
            self._observ.add(node_name)
        elif gvar in self.model.actions:
            self._action.add(node_name)

    def clear_cache(self):
        self._state.clear()
        self._next_state.clear()
        self._interm.clear()
        self._derived.clear()
        self._observ.clear()
        self._action.clear()
        self._gvar_to_node_name.clear()
        self._node_name_to_gvar.clear()
    
    def get_node_name(self, gvar: str, primed: bool=False) -> str:
        if gvar in self._gvar_to_node_name:
            return self._gvar_to_node_name[gvar]

        pvar = self.get_pvar_from_gvar(gvar)
        objects = RDDL2Graph.get_objects(gvar if not primed else gvar[:-1], pvar)
        node_name = f"{pvar}" + ("'" if primed else '') + \
                    f"{('(' + ','.join(objects) + ')') if len(objects) > 0 else ''}"
        self._gvar_to_node_name[gvar] = node_name
        self._node_name_to_gvar[node_name] = gvar
        return node_name
        
    def get_pvar_from_gvar(self, gvar: str):
        pvar = self.model.gvar_to_pvar.get(gvar)
        assert pvar is not None
        return pvar


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--domain", type=str, default='Wildfire')
    parser.add_argument("--instance", type=int, default=0)
    parser.add_argument("--undirected", action='store_true')
    parser.add_argument("--strict_grouping", action='store_true')
    parser.add_argument("--fluent", type=str, default=None)
    parser.add_argument("--gfluent", type=str, default=None)
    parser.add_argument("--simulation", action='store_true')

    args = parser.parse_args()

    r2g = RDDL2Graph(
        domain=args.domain,
        instance=args.instance,
        directed=not args.undirected,
        strict_grouping=args.strict_grouping,
        simulation=args.simulation,
    )
    if args.fluent and args.gfluent:
        r2g.save_dbn(file_name=args.domain, fluent=args.fluent, gfluent=args.gfluent)
    elif args.fluent:
        r2g.save_dbn(file_name=args.domain, fluent=args.fluent)
    else:
        r2g.save_dbn(file_name=args.domain)



from typing import cast, Union, Dict, Any
import numpy as np

from pyRDDLGym.Core.Simulator.RDDLSimulator import RDDLSimulatorWConstraints, RDDLSimulator
from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
from pyRDDLGym.Core.Parser.expr import Expression
from pyRDDLGym.XADD.RDDLModelXADD import RDDLModelWXADD
from pyRDDLGym.XADD.RDDLLevelAnalysisXADD import RDDLLevelAnalysisWXADD


class RDDLSimulatorWXADD(RDDLSimulatorWConstraints):

    def __init__(self, model: PlanningModel,
                 rng: np.random.Generator=np.random.default_rng(),
                 compute_levels: bool=True,
                 max_bound: float=np.inf) -> None:
        self._model = cast(RDDLModelWXADD, model)
        self.context = self._model._context
        self._rng = rng
        
        # perform a dependency analysis and topological sort to compute levels
        dep_analysis = RDDLLevelAnalysisWXADD(self._model)
        self.cpforder = dep_analysis.compute_levels()
        self._order_cpfs = list(sorted(self.cpforder.keys()))
        
        self._action_fluents = set(self._model.actions.keys())
        self._init_actions = self._model.actions.copy()
        
        # non-fluent will never change
        self._subs = self._model.nonfluents.copy()
        
        # is a POMDP
        self._pomdp = bool(self._model.observ)

        # __init__ from 'RDDLSimulatorWConstraints'
        self.epsilon = 0.001
        
        # self.BigM = float(max_bound)
        
        self.BigM = max_bound
        self._bounds = {}
        for state in model.states:
            self._bounds[state] = [-self.BigM, self.BigM]
        for derived in model.derived:
            self._bounds[derived] = [-self.BigM, self.BigM]
        for interm in model.interm:
            self._bounds[interm] = [-self.BigM, self.BigM]
        for action in model.actions:
            self._bounds[action] = [-self.BigM, self.BigM]
        for obs in model.observ:
            self._bounds[obs] = [-self.BigM, self.BigM]

        # actions and states bounds extraction for gym's action and state spaces repots only!
        # currently supports only linear in\equality constraints
        for action_precond in model.preconditions:
            self._parse_bounds_rec(action_precond, self._model.actions)

        for state_inv in model.invariants:
            self._parse_bounds_rec(state_inv, self._model.states)

        for name in self._bounds:
            lb, ub = self._bounds[name]
            RDDLSimulator._check_bounds(
                lb, ub, f'Variable <{name}>', self._bounds[name])
        
    def _sample(self, expr: Union[int, Expression], subs: Dict[str, Any]):
        """Samples the current XADD node by substituting the values stored in 'subs'.

        Args:
            expr (int): The XADD node to which values are substituted
            subs (Dict[str, Any]): The dictionary containing 'var_name -> value' mappings
        """
        if isinstance(expr, Expression):
            return super()._sample(expr, subs)
        else:
            subs = {
                self.var_name_to_sympy_var[self.var_name_to_sympy_var_name.get(v, v)]: val 
                for v, val in subs.items() 
                if not v in self._model.nonfluents 
                and self.var_name_to_sympy_var.get(self.var_name_to_sympy_var_name.get(v, v))
            }
            
            # If there exists a random variable, sample it first
            var_set = self.context.collect_vars(expr)
            rv_set = {v for v in var_set if v in self.context._random_var_set}
            if len(rv_set) > 0:
                for rv in rv_set:
                    if str(rv).startswith('#_UNIFORM'):
                        subs[rv] = self._sample_uniform(expr, subs)
                    elif str(rv).startswith('#_GAUSSIAN'):
                        subs[rv] = self._sample_normal(expr, subs)
                    else:
                        raise ValueError
            cont_assign = {var: val 
                           for var, val in subs.items() 
                           if var in var_set and var in self.context._cont_var_set}
            bool_assign = {var: val 
                           for var, val in subs.items() 
                           if var in var_set and var in self.context._bool_var_set}
            res = self.context.evaluate(node_id=expr,
                                        bool_assign=bool_assign,
                                        cont_assign=cont_assign,
                                        primitive_type=True)
            return res
    
    def _sample_uniform(self, expr, subs):
        return self._rng.uniform(0, 1)
    
    def _sample_normal(self, expr, subs):
        return self._rng.standard_normal()
    
    def _sample_exponential(self, expr, subs):
        raise NotImplementedError
    
    def _sample_bernoulli(self, expr, subs):
        raise NotImplementedError
    
    def _sample_binomial(self, expr, subs):
        raise NotImplementedError
    
    def _sample_gamma(self, expr, subs):
        raise NotImplementedError
    
    def _sample_beta(self, expr, subs):
        raise NotImplementedError
    
    def _sample_weibull(self, expr, subs):
        raise NotImplementedError
    
    def _sample_dirac_delta(self, expr, subs):
        raise NotImplementedError
    
    def _sample_geometric(self, expr, subs):
        raise NotImplementedError
    
    def _sample_gumbel(self, expr, subs):
        raise NotImplementedError
    
    def _sample_kron_delta(self, expr, subs):
        raise NotImplementedError
    
    def _sample_negative_binomial(self, expr, subs):
        raise NotImplementedError
    
    def _sample_poisson(self, expr, subs):
        raise NotImplementedError
    
    def _sample_student(self, expr, subs):
        raise NotImplementedError

    @property
    def sympy_var_name_to_var_name(self):
        return self._model._sympy_var_name_to_var_name

    @property
    def var_name_to_sympy_var_name(self):
        return self._model._var_name_to_sympy_var_name
    
    @property
    def var_name_to_sympy_var(self):
        return self.context._str_var_to_var



from pyRDDLGym.Core.Env import RDDLEnv as RDDLEnv
from pyRDDLGym.Core.Policies.Agents import RandomAgent

# NOTE: set TEMP_VARIANCE = 0 in RDDL to reproduce result
FOLDER = 'pyRDDLGym/Examples/PowerGen/Discrete/'


def power_gen_update(state, action):
    prod_p1 = max(min(action['curProd_p1'], 10), 3)
    prod_p2 = max(min(action['curProd_p2'], 10), 3)
    prod_p3 = max(min(action['curProd_p3'], 10), 3)
    next_prevProd_p1 = prod_p1
    next_prevProd_p2 = prod_p2
    next_prevProd_p3 = prod_p3
    next_temperature = state['temperature']

    demand = 2 + 0.01 * (state['temperature'] - 11.7) ** 2
    fulfilledDemand = min(demand, prod_p1 + prod_p2 + prod_p3)

    reward = -(prod_p1 * 5 + prod_p2 * 5 + prod_p3 * 5) + (fulfilledDemand * 8) \
             - (1000. if (demand > fulfilledDemand) else 0.) \
        + (abs(prod_p1 - state['prevProd_p1']) * 1 + abs(prod_p2 - state['prevProd_p2']) * 1
           + abs(prod_p3 - state['prevProd_p3']) * 1)

    next_interm = {'prod_p1': prod_p1,
                   'prod_p2': prod_p2,
                   'prod_p3': prod_p3,
                   'demand': demand,
                   'fulfilledDemand': fulfilledDemand}

    next_state = {'prevProd_p1': next_prevProd_p1,
                  'prevProd_p2': next_prevProd_p2,
                  'prevProd_p3': next_prevProd_p3,
                  'temperature': next_temperature}

    return reward, next_state, next_interm


def main():
    myEnv = RDDLEnv.RDDLEnv(domain=FOLDER + 'domain.rddl',
                            instance=FOLDER + 'instance0.rddl')
    agent = RandomAgent(action_space=myEnv.action_space,
                        num_actions=myEnv.numConcurrentActions)
    myEnv.model.nonfluents['TEMP_VARIANCE'] = 0.

    from pprint import pprint
    pprint(vars(myEnv.model))

    total_reward = 0
    state = myEnv.reset()
    test_state = state
    for step in range(myEnv.model.horizon):
        myEnv.render()
        action = agent.sample_action()

        next_state, reward, done, info = myEnv.step(action)
        total_reward += reward
        print("\nstep = {}".format(step))
        print('reward = {}'.format(reward))
        print('state = {}'.format(state))
        print('action = {}'.format(action))
        print('next_state = {}'.format(next_state))
        print('interm = {}'.format(myEnv.model.interm))
        state = next_state

        test_reward, next_test_state, test_interm = power_gen_update(
            test_state, action)
        assert test_reward == reward
        assert next_test_state == next_state
        assert test_interm == myEnv.model.interm, str(
            test_interm) + '\n' + str(myEnv.model.interm)
        test_state = next_test_state

    print("episode ended with reward {}".format(total_reward))


if __name__ == "__main__":
    main()



import gym
from gym.spaces import Box, Dict, Discrete, MultiDiscrete
import numpy as np
import os
import pygame
import typing
import warnings

from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLEpisodeAlreadyEndedError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidActionError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLLogFolderError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLNotImplementedError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError

from pyRDDLGym.Core.Compiler.RDDLLiftedModel import RDDLLiftedModel
from pyRDDLGym.Core.Compiler.RDDLValueInitializer import RDDLValueInitializer
from pyRDDLGym.Core.Debug.Logger import Logger, SimLogger
from pyRDDLGym.Core.Env.RDDLConstraints import RDDLConstraints
from pyRDDLGym.Core.Env.RDDLEnvSeeder import RDDLEnvSeeder, RDDLEnvSeederFibonacci
from pyRDDLGym.Core.Parser.parser import RDDLParser
from pyRDDLGym.Core.Parser.RDDLReader import RDDLReader
from pyRDDLGym.Core.Simulator.RDDLSimulator import RDDLSimulator
from pyRDDLGym.Visualizer.ChartViz import ChartVisualizer


def _make_dir(log_path):
    root_path = os.path.dirname(log_path)
    if not os.path.exists(root_path):
        try:
            os.makedirs(root_path)
        except Exception as e:
            if not isinstance(e, FileExistsError):
                raise RDDLLogFolderError(
                    f'Could not create folder at path {root_path}.')
    return log_path

    
class RDDLEnv(gym.Env):
    '''A gym environment class for RDDL domains.'''
    
    @staticmethod
    def build(env_info, env: str, **env_kwargs):
        env = RDDLEnv(domain=env_info.get_domain(),
                      instance=env_info.get_instance(env),
                      **env_kwargs)
        env.set_visualizer(env_info.get_visualizer())
        return env

    def __init__(self, domain: str,
                 instance: str=None,
                 enforce_action_constraints: bool=False,
                 enforce_action_count_non_bool: bool=True,
                 new_gym_api: bool=False,
                 vectorized: bool=False,
                 compact_action_space: bool=False,
                 debug: bool=False,
                 log_path: str=None,
                 backend: RDDLSimulator=RDDLSimulator,
                 backend_kwargs: typing.Dict={},
                 seeds: RDDLEnvSeeder=RDDLEnvSeederFibonacci()):
        '''Creates a new gym environment from the given RDDL domain + instance.
        
        :param domain: the RDDL domain
        :param instance: the RDDL instance
        :param enforce_action_constraints: whether to raise an exception if the
        action constraints are violated
        :param enforce_action_count_non_bool: whether to include non-bool actions
        in check that number of nondef actions don't exceed max-nondef-actions
        :param new_gym_api: whether to use the new Gym API for step()
        :param vectorized: whether actions and states are represented as
        dictionaries of numpy arrays (if True), or as dictionaries of scalars
        :param compact_action_space: whether to use a compact action space most
        suitable for RL implementations such as stable-baselines
        :param debug: whether to log compilation information to a log file
        :param log_path: absolute path to file where simulation log is saved,
        excluding the file extension, None means no logging
        :param backend: the subclass of RDDLSimulator to use as backend for
        simulation (currently supports numpy and Jax)
        :param backend_kwargs: dictionary of additional named arguments to
        pass to backend (must not include logger)
        :param seeds: an instance of RDDLEnvSeeder for generating RNG seeds
        '''
        super(RDDLEnv, self).__init__()
        
        self.domain_text = domain
        self.instance_text = instance
        self.enforce_action_constraints = enforce_action_constraints
        self.enforce_count_non_bool = enforce_action_count_non_bool
        self.new_gym_api = new_gym_api
        self.vectorized = vectorized
        self.compact_action_space = compact_action_space
        
        # vectorized must be true for simplifying action space
        if self.compact_action_space and not self.vectorized:
            warnings.warn('vectorized was set to True because '
                          'compact_action_space was requested.', stacklevel=2)
            self.vectorized = True
        
        # read and parse domain and instance
        reader = RDDLReader(domain, instance)
        domain = reader.rddltxt
        parser = RDDLParser(lexer=None, verbose=False)
        parser.build()
        rddl = parser.parse(domain)
        
        # define the RDDL model
        self.model = RDDLLiftedModel(rddl)
        self.horizon = self.model.horizon
        self.discount = self.model.discount
        self.max_allowed_actions = self.model.max_allowed_actions 
                
        # for logging compilation data
        ast = self.model._AST
        log_fname = f'{ast.domain.name}_{ast.instance.name}'
        logger = Logger(f'{log_fname}_debug.log') if debug else None
        self.logger = logger
        
        # for logging simulation data
        self.simlogger = None
        if log_path is not None and log_path:
            new_log_path = _make_dir(log_path)
            self.simlogger = SimLogger(f'{new_log_path}_log.csv')
            self.simlogger.clear(overwrite=False)
        
        # define the simulation backend  
        self.sampler = backend(self.model,
                               logger=logger,
                               keep_tensors=self.vectorized,
                               **backend_kwargs)
        
        # compute the bounds on fluents from the constraints
        self._bounds = RDDLConstraints(
            self.sampler, vectorized=self.vectorized).bounds
        self._shapes = {var: np.shape(values[0]) 
                        for (var, values) in self._bounds.items()}
        
        # construct the gym observation space
        if self.sampler.isPOMDP:
            statesranges = self.model.observranges
        else:
            statesranges = self.model.statesranges
        if not self.vectorized:
            statesranges = self.model.ground_ranges_from_dict(statesranges)    
        self.observation_space = self._rddl_to_gym_bounds_obs(statesranges)
        
        # construct the gym action space      
        if self.vectorized:
            self._actionsranges = self.model.actionsranges
            self._noop_actions = self.sampler.noop_actions
        else:
            self._actionsranges = self.sampler.grounded_actionsranges
            self._noop_actions = self.sampler.grounded_noop_actions
        self.action_space, self._action_info = self._rddl_to_gym_bounds_act(self._actionsranges)
        
        # set the visualizer
        self._visualizer = ChartVisualizer(self.model)
        self._movie_generator = None
        self.state = None
        self.image = None
        self.window = None
        self.to_render = False
        self.image_size = None
        
        # set roll-out parameters           
        self.trial = 0
        self.currentH = 0
        self.done = False
        self.seeds = iter(seeds)
            
    # ===========================================================================
    # observation and action spaces
    # ===========================================================================
    
    def _rddl_to_gym_bounds_obs(self, ranges):
        result = Dict()
        for (var, prange) in ranges.items():
            
            # enumerated values converted to Discrete space
            if prange in self.model.enums:
                num_objects = len(self.model.objects[prange])
                if self.vectorized:
                    result[var] = Box(0, num_objects - 1,
                                      shape=self._shapes[var],
                                      dtype=np.int32)
                else:
                    result[var] = Discrete(num_objects) 
            
            # real values define a box
            elif prange == 'real':
                low, high = self._bounds[var]
                result[var] = Box(low, high, dtype=np.float32)
            
            # boolean values converted to Discrete space
            elif prange == 'bool':
                if self.vectorized:
                    result[var] = Box(0, 1,
                                      shape=self._shapes[var],
                                      dtype=np.int32)
                else:
                    result[var] = Discrete(2)
            
            # integer values converted to Discrete space
            elif prange == 'int':
                low, high = self._bounds[var]
                low = np.maximum(low, np.iinfo(np.int32).min)
                high = np.minimum(high, np.iinfo(np.int32).max)
                if self.vectorized:
                    result[var] = Box(low, high,
                                      shape=self._shapes[var],
                                      dtype=np.int32)
                else:
                    result[var] = Discrete(int(high - low + 1), start=int(low))
            
            # unknown type
            else:
                raise RDDLTypeError(
                    f'Type <{prange}> of fluent <{var}> is not valid, '
                    f'must be an enumerated or primitive type (real, int, bool).')
        return result
    
    def _rddl_to_gym_bounds_act(self, ranges):
        
        # no simplification rules
        if not self.compact_action_space:
            return self._rddl_to_gym_bounds_obs(ranges), None
        
        # compute whether or not a constraint must be placed on boolean actions
        count_bool = 0
        for (var, prange) in ranges.items():
            if prange == 'bool':
                count_bool += np.prod(self._shapes[var], dtype=np.int64)   
        if self.max_allowed_actions < count_bool:
            if self.max_allowed_actions != 1:
                raise RDDLNotImplementedError(
                    'Simplification of bool action space with max-nondef-actions '
                    'other than 1 or pos-inf is not currently supported.')  
            bool_constraint = True
        else:
            bool_constraint = False
                
        # collect information about action ranges
        locational = {}
        count_disc, count_cont, count_bool = 0, 0, 0
        disc_start, disc_nelem = [], []
        cont_low, cont_high = [], []
        for (var, prange) in ranges.items():
            shape = self._shapes[var]
            num_elements = np.prod(shape, dtype=np.int64)  
            
            # boolean actions without constraint are stored as Discrete
            if prange == 'bool':
                if bool_constraint:
                    locational[var] = ('discrete', (count_bool, num_elements, shape))
                    count_bool += num_elements
                else:
                    disc_start.extend([0] * num_elements)
                    disc_nelem.extend([2] * num_elements)  
                    locational[var] = ('discrete', (count_disc, num_elements, shape))
                    count_disc += num_elements
            
            # integer actions are stored as Discrete
            elif prange == 'int':
                low, high = self._bounds[var]
                low = np.ravel(low, order='C')
                high = np.ravel(high, order='C')
                low = np.maximum(low, np.iinfo(np.int32).min).astype(np.int32)
                high = np.minimum(high, np.iinfo(np.int32).max).astype(np.int32)
                disc_start.extend(low.tolist())
                disc_nelem.extend((high - low + 1).tolist())                
                locational[var] = ('discrete', (count_disc, num_elements, shape))
                count_disc += num_elements
            
            # enum-valued actions are stored as Discrete
            elif prange in self.model.enums:
                num_objects = len(self.model.objects[prange])
                disc_start.extend([0] * num_elements)
                disc_nelem.extend([num_objects] * num_elements)
                locational[var] = ('discrete', (count_disc, num_elements, shape))
                count_disc += num_elements
            
            # real actions are stored as Box
            elif prange == 'real':
                low, high = self._bounds[var]
                low = np.ravel(low, order='C').tolist()
                high = np.ravel(high, order='C').tolist()
                cont_low.extend(low)
                cont_high.extend(high)
                locational[var] = ('continuous', (count_cont, num_elements, shape))
                count_cont += num_elements
            
            # not a valid action type
            else:
                raise RDDLTypeError(
                    f'Type <{prange}> of fluent <{var}> is not valid, '
                    f'must be an enumerated or primitive type (real, int, bool).')
        
        # boolean actions with constraint are stored in the last place instead
        if bool_constraint:
            disc_start.append(0)
            disc_nelem.append(count_bool + 1)
            count_disc += 1
        
        # discrete space
        if len(disc_nelem) == 1:
            disc_space = Discrete(disc_nelem[0])
        elif len(disc_nelem) > 1:
            disc_space = MultiDiscrete(disc_nelem)
        else:
            disc_space = None
        
        # real space
        if count_cont:
            cont_space = Box(np.asarray(cont_low), np.asarray(cont_high), dtype=np.float32)
        else:
            cont_space = None
        
        # simplify space
        combined_space = Dict()
        if disc_space is not None:
            combined_space['discrete'] = disc_space
        if cont_space is not None:
            combined_space['continuous'] = cont_space
        if not combined_space:
            raise RDDLInvalidActionError(
                'RDDL action specification resulted in an empty action space.')
        keys = list(combined_space.keys())
        if len(keys) == 1:
            combined_space = combined_space[keys[0]]
        
        # log information
        if self.logger is not None:
            act_info = '\n\t'.join(
                f'{act}: action_tensor={key}, start={start}, count={count}, shape={shape}'
                for (act, (key, (start, count, shape))) in locational.items())
            bound_info = (f'\tdiscrete_start={disc_start}, discrete_n={disc_nelem}\n'
                          f'\tcontinuous_low={cont_low}, continuous_high={cont_high}')
            self.logger.log(f'[info] computed gym action space:\n' 
                            f'{bound_info}\n'
                            f'\t{act_info}\n'
                            f'[info] final space: {combined_space}\n')
            
        return combined_space, (locational, keys, bool_constraint, np.asarray(disc_start))
    
    def _gym_to_rddl_actions(self, gym_actions):
        
        # no simplification rules
        if not self.compact_action_space:
            return gym_actions
        
        locational, keys, bool_constraint, disc_start = self._action_info
        if len(keys) == 1:
            gym_actions = {keys[0]: gym_actions}  
        
        # process all actions except if active max-nondef-actions constraint
        actions = {}
        for (var, prange) in self._actionsranges.items():
            if not (bool_constraint and prange == 'bool'):
                key, (start, count, shape) = locational[var]
                action_key = gym_actions.get(key, None)
                if action_key is not None:
                    action = np.atleast_1d(action_key)[start:start + count]
                    if key == 'discrete':
                        action = action + disc_start[start:start + count]
                    dtype = RDDLValueInitializer.NUMPY_TYPES.get(
                        prange, RDDLValueInitializer.INT)
                    actions[var] = np.reshape(action, shape, order='C').astype(dtype)
        
        # process the active max-nondef-actions constraint
        action_key = gym_actions.get('discrete', None)
        if bool_constraint and action_key is not None:
            index = np.atleast_1d(action_key)[-1]
            for (var, prange) in self._actionsranges.items():
                if prange == 'bool':
                    _, (start, count, shape) = locational[var]                    
                    index_in_var = index - start
                    if 0 <= index_in_var < count:
                        default_value = self.model.default_values[var]
                        action = np.full(shape=count, fill_value=default_value, dtype=bool)
                        action[index_in_var] ^= True
                        actions[var] = np.reshape(action, shape, order='C')
                        break
            
        return actions
            
    # ===========================================================================
    # core functions
    # ===========================================================================
    
    def seed(self, seed=None):
        self.sampler.seed(seed)
        return [seed]
    
    def set_visualizer(self, viz, movie_gen=None, movie_per_episode=False, **viz_kwargs):
        if viz is not None:
            self._visualizer = viz(self.model, **viz_kwargs)
        self._movie_generator = movie_gen
        self._movie_per_episode = movie_per_episode
        self._movies = 0
        self.to_render = False

    def step(self, actions):
        if self.done:
            raise RDDLEpisodeAlreadyEndedError(
                'The step() function has been called even though the '
                'current episode has terminated: please call reset().')            
            
        actions = self._gym_to_rddl_actions(actions)
        
        # cast non-boolean actions to boolean
        fixed_actions = self._noop_actions.copy()
        for (var, values) in actions.items():
            if self._actionsranges.get(var, '') == 'bool':
                if np.shape(values):
                    fixed_actions[var] = np.asarray(values, dtype=bool)
                else:
                    fixed_actions[var] = bool(values)
            else:
                fixed_actions[var] = values
        actions = fixed_actions

        # check action constraints
        sampler = self.sampler
        sampler.check_default_action_count(actions, self.enforce_count_non_bool)
        if self.enforce_action_constraints:
            sampler.check_action_preconditions(actions, silent=False)
        
        # sample next state and reward
        obs, reward, self.done = sampler.step(actions)
        self.state = sampler.states
            
        # check if the state invariants are satisfied
        if self.done:
            out_of_bounds = False
        else:
            out_of_bounds = not sampler.check_state_invariants(silent=True)
            
        # log to file
        if self.simlogger is not None:
            if self.vectorized:
                log_obs = self.model.ground_values_from_dict(obs)
                log_action = self.model.ground_values_from_dict(actions)
            else:
                log_obs = obs
                log_action = actions
            self.simlogger.log(log_obs, log_action, reward, self.done, self.currentH)
        
        # update step horizon
        self.currentH += 1
        if self.currentH == self.horizon:
            self.done = True
        
        # produce array outputs for vectorized option
        if self.vectorized:
            obs = {var: np.atleast_1d(value) for (var, value) in obs.items()}
            
        if self.new_gym_api:
            return obs, reward, self.done, out_of_bounds, {}
        else:
            return obs, reward, self.done, {}

    def reset(self, seed=None, options=None):
        
        # reset counters and internal state
        sampler = self.sampler
        obs, self.done = sampler.reset()
        self.state = sampler.states
        self.trial += 1
        self.currentH = 0
        
        # update movie generator
        if self._movie_generator is not None and self._visualizer is not None:
            if self.vectorized:
                state = self.model.ground_values_from_dict(self.state)
            else:
                state = self.state
                
            image = self._visualizer.render(state)
            self.image_size = image.size
            if self._movie_per_episode:
                self._movie_generator.save_animation(
                    self._movie_generator.env_name + '_' + str(self._movies))
                self._movies += 1
            self._movie_generator.save_frame(image)            
        
        # update random generator seed
        if seed is not None:
            self.seed(seed)
        else:
            seed = next(self.seeds)
            if seed is not None:
                self.seed(seed)

        # logging
        if self.simlogger:
            text = (f'######################################################\n'
                    f'New Trial, seed={seed}\n'
                    f'######################################################')
            self.simlogger.log_free(text)
            
        # produce array outputs for vectorized option
        if self.vectorized:
            obs = {var: np.atleast_1d(value) for (var, value) in obs.items()}
            
        if self.new_gym_api:
            return obs, {}
        else:
            return obs

    def pilImageToSurface(self, pilImage):
        return pygame.image.fromstring(
            pilImage.tobytes(), pilImage.size, pilImage.mode).convert()

    def render(self, to_display=True):
        image = None
        if self._visualizer is not None:
            if self.vectorized:
                state = self.model.ground_values_from_dict(self.state)
            else:
                state = self.state
            
            # update the screen
            image = self._visualizer.render(state)
            self.image_size = image.size
            if to_display:
                if not self.to_render:
                    self.to_render = True
                    pygame.init()
                    self.window = pygame.display.set_mode(
                        (self.image_size[0], self.image_size[1]))
                self.window.fill(0)
                pygameSurface = self.pilImageToSurface(image)
                self.window.blit(pygameSurface, (0, 0))
                pygame.display.set_caption(self.model.instanceName())
                pygame.display.flip()
                
                # prevents the window from freezing up midway
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        pass
            
            # capture frame to disk
            if self._movie_generator is not None:
                self._movie_generator.save_frame(image)
    
        return image
    
    def close(self):
        if self.simlogger:
            self.simlogger.close()
        
        # close rendering and save animation  
        if self.to_render:
            pygame.display.quit()
            pygame.quit()    
            
            # prepare the animation from captured frames
            if self._movie_generator is not None:
                self._movie_generator.save_animation(
                    self._movie_generator.env_name + '_' + str(self._movies))
                self._movies += 1



import json
import numpy as np
import os
import sys

from pyRDDLGym.Core.Policies.RDDLSimAgent import RDDLSimAgent

# read the command line args
args = sys.argv[1:]
if len(args) != 1:
    print('')
    print('    Usage: python rddlsim.py <rounds>')
    print(f'    Given: {args}')
    print('')
    sys.exit(1)
rounds = int(args[0])

# load the RDDL files
RDDL_path = os.environ.get('RDDL')
domain_path = os.path.join(RDDL_path, 'domain.rddl')
instance_path = os.path.join(RDDL_path, 'instance.rddl')

# launch the RDDL server
agent = RDDLSimAgent(domain_path, instance_path, rounds, 99999)
agent.run()

# aggregate results
round_returns = [float(round_data[-1]['reward']) for round_data in agent.logs]
stats = {
    'mean': np.mean(round_returns),
    'median': np.median(round_returns),
    'min': np.min(round_returns),
    'max': np.max(round_returns),
    'std': np.std(round_returns)
}

# save outputs
out_path = os.environ.get('PROST_OUT')
dom_name = agent.env.model._AST.domain.name
inst_name = agent.env.model._AST.domain.name
with open(os.path.join(out_path, f'summary_{dom_name}_{inst_name}.json'), 'w') as f:
    json.dump(stats, f)    
agent.dump_data(os.path.join(out_path, f'data_{dom_name}_{inst_name}.json'))



from pyRDDLGym.Core.Parser import parser as parser
from pyRDDLGym.Core.Parser import RDDLReader as RDDLReader
from pyRDDLGym.Core import Grounder as RDDLGrounder
from pyRDDLGym.Core.Simulator.RDDLSimulator import RDDLSimulator

DOMAIN = 'power_unit_commitment.rddl'

# DOMAIN = 'ThiagosReservoir.rddl'
# DOMAIN = 'Thiagos_Mars_Rover.rddl'
# DOMAIN = 'dbn_prop.rddl'
# DOMAIN = 'RamMod_Thiagos_HVAC.rddl'
DOMAIN = 'RamMod_smaller_Thiagos_HVAC.rddl'
# DOMAIN = 'RamMod_Thiagos_HVAC_grounded.rddl'

# DOMAIN = 'Thiagos_HVAC.rddl'
# DOMAIN = 'Thiagos_HVAC_grounded.rddl'
# DOMAIN = 'wildfire_mdp.rddl'
DOMAIN = 'recsim_ecosystem_welfare.rddl'

def main():

    MyReader = RDDLReader.RDDLReader('RDDL/' + DOMAIN)
    # MyReader = RDDLReader.RDDLReader('RDDL/power_unit_commitment_domain.rddl',
    #                                  'RDDL/power_unit_commitment_instance.rddl')
    domain = MyReader.rddltxt

    MyLexer = parser.RDDLlex()
    MyLexer.build()
    MyLexer.input(domain)
    # [token for token in MyLexer._lexer]

    # build parser - built in lexer, non verbose
    MyRDDLParser = parser.RDDLParser(None, False)
    MyRDDLParser.build()

    # parse RDDL file
    rddl_ast = MyRDDLParser.parse(domain)

    # MyXADDTranslator = XADDTranslator(rddl_ast)
    # MyXADDTranslator.Translate()

    grounder = RDDLGrounder.RDDLGrounder(rddl_ast)
    model = grounder.Ground()
    # print(model.states)
    # print(model.init_state)
    # for cpf_key, primed_cpf in model.interm.items():
    #     expr = model.cpfs[primed_cpf]
    #     print(primed_cpf, expr)
    # print(model.cpfs)
    # pprint(vars(model))
    # #
    # good_policy = True
    sampler = RDDLSimulator(model)
    loops = 1
    #
    print('\nstarting simulation')
    for h in range(loops):
        state = sampler.reset_state()
        print(state)
        action = {
        'recommend(c1, i1)': True,
        'recommend(c2, i2)': True,
        'recommend(c3, i3)': True,
        'recommend(c4, i4)': True,
        'recommend(c5, i5)': True,
        }
        state = sampler.sample_next_state(action)
        print(state)
    #     total_reward = 0.
    #     for _ in range(2000):
    #         sampler.check_state_invariants()
    #         sampler.check_action_preconditions()
    #         actions = {'AIR_r1': 0., 'AIR_r2': 0., 'AIR_r3': 0.}
    #         if good_policy:
    #             if state['TEMP_r1'] < 20.5:
    #                 actions['AIR_r1'] = 5.
    #             if state['TEMP_r2'] < 20.5:
    #                 actions['AIR_r2'] = 5.
    #             if state['TEMP_r3'] < 20.5:
    #                 actions['AIR_r3'] = 5.
    #             if state['TEMP_r1'] > 23.:
    #                 actions['AIR_r1'] = 0.
    #             if state['TEMP_r2'] > 23.:
    #                 actions['AIR_r2'] = 0.
    #             if state['TEMP_r3'] > 23.:
    #                 actions['AIR_r3'] = 0.
    #         state = sampler.sample_next_state(actions)
    #         reward = sampler.sample_reward()
    #         sampler.update_state()
    #         if h == 0:
    #             print('state = {}'.format(state))
    #             print('reward = {}'.format(reward))
    #             print('derived = {}'.format(model.derived))
    #             print('interm = {}'.format(model.interm))
    #             print('')
    #         total_reward += reward
    #     print('trial {}, total reward {}'.format(h, total_reward))

    # grounder = RDDLGrounder.RDDLGrounder(rddl_ast)
    # grounder.Ground()
    # pprint(vars(grounder))

    # grounder.InitGround()
    # generator = RDDLGenerator.RDDLGenerator(rddl_ast)
    # rddl = generator.GenerateRDDL()


    print("reached end of test.py")

if __name__ == "__main__":
    main()




# from Parser import parser as parser
# from Parser import RDDLReader as RDDLReader
# import Grounder.RDDLGrounder as RDDLGrounder
# from Visualizer.MarsRoverDisplay import MarsRoverDisplay
# from Visualizer.ReservoirDisplay import ReservoirDisplay
# from Visualizer.HVACDisplay import HVACDisplay

# DOMAIN = 'power_unit_commitment.rddl'

# # DOMAIN = 'ThiagosReservoir.rddl'
# DOMAIN = 'ThiagosReservoir_grounded.rddl'
# # DOMAIN = 'Thiagos_Mars_Rover.rddl'
# # DOMAIN = 'Thiagos_Mars_Rover_grounded.rddl'
# # DOMAIN = 'Thiagos_HVAC.rddl'
# # DOMAIN = 'dbn_prop.rddl'
# # DOMAIN = 'Thiagos_HVAC_grounded.rddl'
# # DOMAIN = 'wildfire_mdp.rddl'
# DOMAIN = 'recsim_ecosystem_welfare.rddl'

# def main():

#     MyReader = RDDLReader.RDDLReader('RDDL/' + DOMAIN)
#     # MyReader = RDDLReader.RDDLReader('RDDL/power_unit_commitment_domain.rddl',
#     #                                  'RDDL/power_unit_commitment_instance.rddl')
#     domain = MyReader.rddltxt

#     MyLexer = parser.RDDLlex()
#     MyLexer.build()
#     MyLexer.input(domain)
#     # [token for token in MyLexer._lexer]

#     # build parser - built in lexer, non verbose
#     MyRDDLParser = parser.RDDLParser(None, False)
#     MyRDDLParser.build()

#     # parse RDDL file
#     rddl_ast = MyRDDLParser.parse(domain)



#     # MyXADDTranslator = XADDTranslator(rddl_ast)
#     # MyXADDTranslator.Translate()

#     from pprint import pprint
#     grounder = RDDLGrounder.RDDLGrounder(rddl_ast)
#     model = grounder.Ground()
#     # print(model.states)
#     # print(model.init_state)
#     # for cpf_key, primed_cpf in model.interm.items():
#     #     expr = model.cpfs[primed_cpf]
#     #     print(primed_cpf, expr)
#     # print(model.cpfs)
#     # pprint(vars(model))
#     # #
#     # good_policy = True
#     sampler = RDDLSimulator(model)
#     loops = 1
#     #
#     print('\nstarting simulation')
#     for h in range(loops):
#         state = sampler.reset_state()
#         print(state)
#         action = {
#         'recommend(c1, i1)': True,
#         'recommend(c2, i2)': True,
#         'recommend(c3, i3)': True,
#         'recommend(c4, i4)': True,
#         'recommend(c5, i5)': True,
#         }
#         state = sampler.sample_next_state(action)
#         print(state)
#     #     total_reward = 0.
#     #     for _ in range(2000):
#     #         sampler.check_state_invariants()
#     #         sampler.check_action_preconditions()
#     #         actions = {'AIR_r1': 0., 'AIR_r2': 0., 'AIR_r3': 0.}
#     #         if good_policy:
#     #             if state['TEMP_r1'] < 20.5:
#     #                 actions['AIR_r1'] = 5.
#     #             if state['TEMP_r2'] < 20.5:
#     #                 actions['AIR_r2'] = 5.
#     #             if state['TEMP_r3'] < 20.5:
#     #                 actions['AIR_r3'] = 5.
#     #             if state['TEMP_r1'] > 23.:
#     #                 actions['AIR_r1'] = 0.
#     #             if state['TEMP_r2'] > 23.:
#     #                 actions['AIR_r2'] = 0.
#     #             if state['TEMP_r3'] > 23.:
#     #                 actions['AIR_r3'] = 0.
#     #         state = sampler.sample_next_state(actions)
#     #         reward = sampler.sample_reward()
#     #         sampler.update_state()
#     #         if h == 0:
#     #             print('state = {}'.format(state))
#     #             print('reward = {}'.format(reward))
#     #             print('derived = {}'.format(model.derived))
#     #             print('interm = {}'.format(model.interm))
#     #             print('')
#     #         total_reward += reward
#     #     print('trial {}, total reward {}'.format(h, total_reward))

#     grounder = RDDLGrounder.RDDLGroundedGrounder(rddl_ast)
#     model = grounder.Ground()
#     # marsVisual = MarsRoverDisplay(model, grid_size=[50,50], resolution=[500,500])
#     # marsVisual.display_img(duration=0.5)
#     # marsVisual.save_img('./pict2.png')

#     # reservoirVisual = ReservoirDisplay(model, grid_size=[50,50], resolution=[500,500])

#     HVACVisual = HVACDisplay(model, grid_size=[50,50], resolution=[500,500])


#     # print(model._nonfluents)
#     # print(model._states)
#     # print(model._objects)



#     # generator = RDDLGenerator.RDDLGenerator(rddl_ast)
#     # rddl = generator.GenerateRDDL()
#     # print(rddl)

#     print("reached end of test.py")

# if __name__ == "__main__":
#     main()



# GROUNDER OLD
# class RDDLGroundedGrounder(Grounder):
#
#   def __init__(self, RDDL_AST):
#     super(RDDLGroundedGrounder, self).__init__()
#     self._AST = RDDL_AST
#     # self._Model = None
#     self._actions = {}
#     self._nonfluents = {}
#     self._states = {}
#     self._nextstates = {}
#     self._prevstates = {}
#     self._init_state = {}
#     self._cpfs = {}
#     self._cpforder = {0: []}
#     self._derived = {}
#     self._interm = {}
#     self._objects = {}
#
#     self._reward = None
#     self._preconditions = []
#     self._invariants = []
#
#     self._actionsranges = {}
#     self._statesranges = {}
#
#   def Ground(self):
#     # there are no objects or types in grounded domains
#
#     # initialize the Model object
#     model = RDDLModel()
#
#     self._getObjects()
#
#     # ground pvariables and appropriate cpfs if applicable
#     # update pvariables
#     self._groundPvariables()
#
#     # update non_fluents values in case the default values were overridden in the instance
#     self._groundNonfluents()
#
#     # ground init_state
#     self._groundInitState()
#
#     # ground reward
#     self._groundReward()
#
#     # ground constraints
#     self._groundPreConstraints()
#
#     # update model object
#     model.states = self._states
#     model.actions = self._actions
#     model.nonfluents = self._nonfluents
#     model.next_state = self._nextstates
#     model.prev_state = self._prevstates
#     model.init_state = self._init_state
#     model.cpfs = self._cpfs
#     model.cpforder = self._cpforder
#     model.reward = self._reward
#     model.preconditions = self._preconditions
#     model.invariants = self._invariants
#     model.derived = self._derived
#     model.interm = self._interm
#
#     # new properties
#     model.max_allowed_actions = self._groundMaxActions()
#     model.horizon = self._groundHorizon()
#     model.discount = self._groundDiscount()
#     model.actionsranges = self._actionsranges
#     model.statesranges = self._statesranges
#     model.objects = self._objects
#     # new properties
#
#     return model
#
#   def _getObjects(self):
#     self._objects = {}
#     try:
#       for type in self._AST.non_fluents.objects:
#         self._objects[type[0]] = type[1]
#     except:
#       return
#
#   def _groundHorizon(self):
#     return self._AST.instance.horizon
#
#   def _groundMaxActions(self):
#     numactions = self._AST.instance.max_nondef_actions
#     if numactions == 'pos-inf':
#       return len(self._actions)
#     else:
#       return int(numactions)
#
#   def _groundDiscount(self):
#     return self._AST.instance.discount
#
#   def _groundPvariables(self):
#     for pvariable in self._AST.domain.pvariables:
#       name = pvariable.name
#       if pvariable.fluent_type == 'non-fluent':
#         self._nonfluents[name] = pvariable.default
#       elif pvariable.fluent_type == 'action-fluent':
#         self._actions[name] = pvariable.default
#         self._actionsranges[name] = pvariable.range
#       elif pvariable.fluent_type == 'state-fluent':
#         cpf = None
#         next_state = name + '\''
#         for cpfs in self._AST.domain.cpfs[1]:
#           if cpfs.pvar[1][0] == next_state:
#             cpf = cpfs
#         if cpf is not None:
#           self._states[name] = pvariable.default
#           self._statesranges[name] = pvariable.range
#           self._nextstates[name] = next_state
#           self._prevstates[next_state] = name
#           self._cpfs[next_state] = cpf.expr
#           self._cpforder[0].append(name)
#       elif pvariable.fluent_type == 'derived-fluent':
#         cpf = None
#         for cpfs in self._AST.domain.derived_cpfs:
#           if cpfs.pvar[1][0] == name:
#             cpf = cpfs
#         if cpf is not None:
#           self._derived[name] = pvariable.default
#           self._cpfs[name] = cpf.expr  #sim expects expression here.
#           level = pvariable.level
#           if level is None:
#             level = 1
#           if level in self._cpforder:
#             self._cpforder[level].append(name)
#           else:
#             self._cpforder[level] = [name]
#       elif pvariable.fluent_type == 'interm-fluent':
#         cpf = None
#         for cpfs in self._AST.domain.intermediate_cpfs:
#           if cpfs.pvar[1][0] == name:
#             cpf = cpfs
#         if cpf is not None:
#           self._interm[name] = pvariable.default
#           self._cpfs[name] = cpf.expr
#           level = pvariable.level
#           if level is None:
#             level = 1
#           if level in self._cpforder:
#             self._cpforder[level].append(name)
#           else:
#             self._cpforder[level] = [name]
#
#   def _groundNonfluents(self):
#     if hasattr(self._AST.non_fluents, 'init_non_fluent'):
#       for init_vals in self._AST.non_fluents.init_non_fluent:
#         key = init_vals[0][0]
#         val = init_vals[1]
#         self._nonfluents[key] = val
#
#   def _groundInitState(self):
#     self._init_state = self._states.copy()
#     if hasattr(self._AST.instance, 'init_state'):
#       for init_vals in self._AST.instance.init_state:
#         key = init_vals[0][0]
#         val = init_vals[1]
#         self._init_state[key] = val
#
#   def _groundReward(self):
#     self._reward = self._AST.domain.reward
#
#   def _groundPreConstraints(self):
#     if hasattr(self._AST.domain, 'preconds'):
#       for precond in self._AST.domain.preconds:
#         self._preconditions.append(precond)
#
#     if hasattr(self._AST.domain, 'invariants'):
#       for inv in self._AST.domain.invariants:
#         self._invariants.append(inv)


# import os
# import sys
#
# from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
# from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import load_config
# from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxRDDLBackpropPlanner
# from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxOfflineController
# from pyRDDLGym.Examples.ExampleManager import ExampleManager
#
#
# def main(domain, instance):
#
#     # set up the environment
#     info = ExampleManager.GetEnvInfo(domain)    
#     env = RDDLEnv.build(info, instance, enforce_action_constraints=True)
#
#     # load the config file with planner settings
#     abs_path = os.path.dirname(os.path.abspath(__file__))
#     config_path = os.path.join(abs_path, 'JaxPlanConfigs', f'{domain}_slp.cfg') 
#     planner_args, _, train_args = load_config(config_path)
#
#     # create the planning algorithm and controller
#     planner = JaxRDDLBackpropPlanner(rddl=env.model, **planner_args)
#     controller = JaxOfflineController(planner, params={}, **train_args)
#
#     # expand budget in config
#     train_args['train_seconds'] = 60
#
#     # train for 10 seconds, evaluate, then repeat
#     eval_period = 10
#     time_last_eval = 0
#     for callback in planner.optimize_generator(**train_args):
#         if callback['elapsed_time'] - time_last_eval > eval_period:
#             controller.params = callback['best_params']
#             controller.evaluate(env, verbose=False, render=True)
#             time_last_eval = callback['elapsed_time']
#
#     env.close()
#
# if __name__ == "__main__":
#     args = sys.argv[1:]
#     if len(args) < 2:
#         print('python JaxExample2.py <domain> <instance>')
#         exit(0)
#     domain, instance = args[:2]
#     main(domain, instance)
#




import os
import random
from typing import Dict

from pyRDDLGym.Examples.InstanceGenerator import InstanceGenerator


class UAVInstanceGenerator(InstanceGenerator):
    
    def get_env_path(self) -> str:
        return os.path.join('UAV', 'Continuous')
    
    def get_domain_name(self) -> str:
        return 'kinematic_UAVs_con'
    
    def sample_instance(self, params: Dict[str, object]) -> Dict[str, object]:
        aircraft = [f'a{i + 1}' for i in range(params['num_aircraft'])]
        
        nonfluents = {}
        xrange = params['xrange']
        yrange = params['yrange']
        zrange = params['zrange']
        for ac in aircraft:
            nonfluents[f'GOAL-X({ac})'] = random.uniform(*xrange)
            nonfluents[f'GOAL-Y({ac})'] = random.uniform(*yrange)
            nonfluents[f'GOAL-Z({ac})'] = random.uniform(*zrange)
            nonfluents[f'MIN-ACC({ac})'] = -10.0
            nonfluents[f'MAX-ACC({ac})'] = 10.0            
        nonfluents['RANDOM-WALK-COEFF'] = params['variance']
        
        crafts = list(range(len(aircraft)))
        random.shuffle(crafts)
        controlled = crafts[:params['num_control']]
        controlled = sorted(controlled)
        for ac in controlled:
            nonfluents[f'CONTROLLABLE({aircraft[ac]})'] = True
        
        states = {}
        for ac in aircraft:
            states[f'pos-x({ac})'] = random.uniform(*xrange)
            states[f'pos-y({ac})'] = random.uniform(*yrange)
            states[f'pos-z({ac})'] = zrange[0]
                        
        return {
            'objects': {'aircraft': aircraft},
            'non-fluents': nonfluents,
            'init-states': states,
            'horizon': params['horizon'],
            'discount': params['discount'],
            'max-nondef-actions': 'pos-inf'
        }




import jax
import jax.numpy as jnp
import numpy as np
import os
import plotly.graph_objects as go

from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import load_config, JaxRDDLBackpropPlanner
from pyRDDLGym.Examples.ExampleManager import ExampleManager


def sample_points(center, dir1, dir2, x1min, x1max, x2min, x2max, n):
    x1 = np.linspace(x1min, x1max, n)
    x2 = np.linspace(x2min, x2max, n)
    grid = np.meshgrid(x1, x2)
    dirs = np.column_stack((dir1, dir2))
    points = np.einsum('ij,jkl->ikl', dirs, np.array(grid))
    points = np.reshape(points, newshape=(points.shape[0], -1)).T
    points += center
    return points, (x1, x2)


def sample_directions(dim, loc=0.0, scale=1.0):
    dir1 = np.random.normal(loc=loc, scale=scale, size=(dim,))
    dir1 = dir1 / np.linalg.norm(dir1)
    dir2 = np.random.normal(loc=loc, scale=scale, size=(dim,))
    dir2 = dir2 / np.linalg.norm(dir2)
    return dir1, dir2        


def _shape_info(planner):
    shapes, starts, sizes = {}, {}, {}
    istart = 0
    for name in planner.compiled.rddl.actions:
        shapes[name] = (planner.horizon,) + \
            np.shape(planner.compiled.init_values[name])
        starts[name] = istart
        sizes[name] = np.prod(shapes[name], dtype=int)
        istart += sizes[name]    
    count_entries = sum(sizes.values())
    return shapes, starts, sizes, count_entries

    
def loss_surface(planner, train_args, w=None, wa=None, train=True):
    
    model_params = planner.compiled.model_params
    if w is not None:
        model_params = {name: w for name in model_params}
        
    hyperparams = train_args.get('policy_hyperparams', {})
    if wa is not None:
        hyperparams = {name: wa for name in hyperparams}
    
    # plan vector
    shapes, starts, sizes, _ = _shape_info(planner)
    
    @jax.jit
    def unravel_params(params):
        policy_params = {}
        for name in planner.compiled.rddl.actions:
            start = starts[name]
            policy_params[name] = jnp.reshape(
                params[start:start + sizes[name]], newshape=shapes[name])
        return policy_params
    
    # loss surface
    key = train_args['key']
    train_subs, test_subs = planner._batched_init_subs(
        planner.compiled.init_values if train else planner.test_compiled.init_values)
    subs = train_subs if train else test_subs
    loss_fn = planner.train_loss if train else planner.test_loss
    
    @jax.jit
    def loss_func(params):
        params = unravel_params(params)
        loss, _ = loss_fn(key, params, hyperparams, subs, model_params)
        return loss
        
    @jax.jit
    def loss_func_batched(params):
        return jnp.ravel(jax.vmap(loss_func)(params))
    
    return loss_func_batched


def plot_surface(x, y, z, name):
    z = np.reshape(z, newshape=(x.size, y.size))
    fig = go.Figure(data=[go.Surface(z=z, x=x, y=y)])
    fig.update_layout(title='Loss surface', autosize=False, width=1000, height=1000)
    fig.write_html(f'{name}.html')
    fig.show()


def run_experiment(config_path, dom, env,
                   xmax=25.0, n=500, iters=100,
                   ws=[(100.0, 5.0), (100.0, 5.0), (10000.0, 100.0)]):
    
    # solve with default parameters
    planner_args, _, train_args = load_config(config_path)
    planner = JaxRDDLBackpropPlanner(rddl=env.model, **planner_args)
    params = planner.optimize(**train_args)
    
    # reshape to solution vector
    _, starts, sizes, count_entries = _shape_info(planner)
    center = np.zeros((count_entries,))
    for name, value in params.items():
        start = starts[name]
        center[start:start + sizes[name]] = np.ravel(value)
    
    # sample evaluation points and directions
    samples = n ** 2
    bs = samples // iters
    d1, d2 = sample_directions(count_entries)
    points, (x, y) = sample_points(center, d1, d2, -xmax, xmax, -xmax, xmax, n)
    
    # evaluate loss surfaces
    for i, (w, wa) in enumerate(ws):
        loss_func = loss_surface(planner, train_args, w=w, wa=wa, train=i > 0)
        zbatches = []
        for j in range(iters):
            print(f'batch {j}')
            batch = jnp.asarray(points[j * bs:j * bs + bs,:])
            zbatch = np.asarray(loss_func(batch))
            zbatches.append(zbatch)
        z = np.concatenate(zbatches)
        plot_surface(x, y, z, f'surface_{i}_{dom}')


if __name__ == '__main__': 
    EnvInfo = ExampleManager.GetEnvInfo('Wildfire')    
    env = RDDLEnv(domain=EnvInfo.get_domain(),
                  instance=EnvInfo.get_instance(0))
    
    abs_path = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(abs_path, 'JaxPlanConfigs', 'Wildfire_slp.cfg')
    
    run_experiment(config_path, 'Wildfire', env)



import re

from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLParseError


class RDDLReader(object):

    comment = r'\/\/.*?\n'
    comment_ws = r'(\s*?\n\s*?)+'
    domain_block = r'(?s)domain.*?\{.*?pvariables.*?cpfs.*?reward.*?;.*?\}[^;]'
    nonfluent_in_domain = r'(?s)\{\s*?non-fluent,.*?\};'
    nonfluent_block = r'(?s)non-fluents[^=]*?\{.*?\}[^;]'
    instance_block = r'(?s)instance.*?\{.*\}[^;]'

    def __init__(self, dom, inst=None):
        with open(dom) as file:
            dom_txt = file.read()
        dom_txt = self._removeComments(dom_txt)
        dom_txt = dom_txt + '\n'

        if inst is not None:
            with open(inst) as file:
                inst_txt = file.read()
            inst_txt = self._removeComments(inst_txt)
            dom_txt = dom_txt + '\n\n' + inst_txt + '\n'

        # inspect rddl if three block are present - domain, non-fluent, instance
        m = re.search(self.domain_block, dom_txt)
        if m is None:
            raise RDDLParseError('domain {...} block is missing or contains a syntax error.')

        domaintxt = m.group(0)
        m = re.search(self.nonfluent_in_domain, domaintxt)
        if m is not None:
            m = re.search(self.nonfluent_block, dom_txt)
            if m is None:
                raise RDDLParseError('non-fluents {...} block is missing or contains a syntax error.')

        m = re.search(self.instance_block, dom_txt)
        if m is None:
            raise RDDLParseError('instance {...} block is missing or contains a syntax error.')

        self.dom_txt = dom_txt

    @property
    def rddltxt(self):
        return self.dom_txt

    def _removeComments(self, txt):
        txt = re.sub(self.comment, '\n', txt)
        txt = re.sub(self.comment_ws, '\n', txt)
        return txt



from typing import Union, cast, Dict, Set
from xaddpy.xadd import XADD

from pyRDDLGym.Core.Compiler.RDDLLevelAnalysis import RDDLLevelAnalysis
from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
from pyRDDLGym.XADD.RDDLModelXADD import RDDLModelWXADD
from pyRDDLGym.Core.Parser.expr import Expression


class RDDLLevelAnalysisWXADD(RDDLLevelAnalysis):
    
    def __init__(self, model: PlanningModel, 
                 disallow_state_synchrony: bool = False) -> None:
        super().__init__(model, disallow_state_synchrony)
        self._model = cast(RDDLModelWXADD, self._model)
        self._context: XADD = self._model._context
    
    def _update_call_graph(self, graph: Dict[str, Set[str]], 
                           cpf: str, 
                           expr: Union[int, Expression]):
        if isinstance(expr, Expression):
            super()._update_call_graph(graph, cpf, expr)
        else:
            var_set = self._context.collect_vars(expr)
            for v in var_set:
                v_name = self._model._sympy_var_name_to_var_name.get(str(v), str(v))
                if v_name not in self._model.nonfluents and v not in self._model.rvs:
                    self._check_is_fluent(cpf, v_name)
                    graph.setdefault(cpf, set()).add(v_name)
    
    # dependency analysis
    def build_call_graph(self) -> Dict[str, Set[str]]:
        
        # compute call graph of CPs and check validity
        cpf_graph = {}
        for cpf, expr in self._model.cpfs.items():
            self._update_call_graph(cpf_graph, cpf, expr)
            if cpf not in cpf_graph:
                cpf_graph[cpf] = set()
        self._check_deps_by_fluent_type(cpf_graph)
        
        # check validity of reward, constraints, termination
        for name, exprs in [
            ('reward', [self._model.reward]),
            ('action precondition', self._model.preconditions),
            ('state invariant', self._model.invariants),
            ('termination', self._model.terminals)
        ]:
            call_graph = {}
            for expr in exprs:
                if isinstance(expr, int):
                    self._update_call_graph(call_graph, name, expr)
                else:
                    super()._update_call_graph(call_graph, name, expr)
            self._check_deps_by_fluent_type(call_graph)
            
        return cpf_graph



# This file is based on thiago pbueno's pyrddl.
# https://github.com/thiagopbueno/pyrddl
# it was adapted and extended for pyRDDLGym

def rename_next_state_fluent(name: str) -> str:
    '''Returns next state fluent canonical name.
    Args:
        name (str): The current state fluent name.
    Returns:
        str: The next state fluent name.
    '''
    i = name.index('/')
    functor = name[:i-1]
    arity = name[i+1:]
    return "{}/{}".format(functor, arity)


def rename_state_fluent(name: str) -> str:
    '''Returns current state fluent canonical name.
    Args:
        name (str): The next state fluent name.
    Returns:
        str: The current state fluent name.
    '''
    i = name.index('/')
    functor = name[:i]
    arity = name[i+1:]
    return "{}'/{}".format(functor, arity)



from abc import ABCMeta
import itertools
import numpy as np
from typing import Dict, Iterable, List, Tuple

from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidObjectError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLUndefinedVariableError

from pyRDDLGym.Core.Parser.expr import Expression, Value


class PlanningModel(metaclass=ABCMeta):
    '''The base class representing all RDDL domain + instance.
    '''
    
    FLUENT_SEP = '___'
    OBJECT_SEP = '__'
    NEXT_STATE_SYM = '\''

    def __init__(self):
        self._AST = None
        self._nonfluents = None
        self._states = None
        self._statesranges = None
        self._nextstates = None
        self._prevstates = None
        self._initstate = None
        self._actions = None
        self._objects = None
        self._objects_rev = None
        self._enums = None
        self._actionsranges = None
        self._derived = None
        self._interm = None
        self._observ = None
        self._observranges = None
        self._cpfs = None
        self._cpforder = None
        self._gvar_to_cpforder = None
        self._reward = None
        self._terminals = None
        self._preconditions = None
        self._invariants = None
        self._gvar_to_pvar = None
        self._pvar_to_type = None
        self._gvar_to_type = None

        self._max_allowed_actions = None
        self._horizon = None
        self._discount = None
        
        self._param_types = None
        self._variable_types = None
        self._variable_ranges = None
        
        self._grounded_names = None
        self._index_of_object = None
        self._default_values = None
        
    # ===========================================================================
    # properties
    # ===========================================================================
    
    def SetAST(self, AST):
        self._AST = AST
    
    def domainName(self):
        if self._AST is None:
            return None
        else:
            return self._AST.domain.name
    
    def instanceName(self):
        if self._AST is None:
            return None
        else:
            return self._AST.instance.name
        
    @property
    def objects(self):
        return self._objects

    @objects.setter
    def objects(self, value):
        self._objects = value
    
    @property
    def objects_rev(self):
        return self._objects_rev

    @objects_rev.setter
    def objects_rev(self, value):
        self._objects_rev = value
    
    @property
    def enums(self):
        return self._enums

    @enums.setter
    def enums(self, value):
        self._enums = value
    
    @property
    def nonfluents(self):
        return self._nonfluents

    @nonfluents.setter
    def nonfluents(self, val):
        self._nonfluents = val

    @property
    def states(self):
        return self._states

    @states.setter
    def states(self, val):
        self._states = val

    @property
    def statesranges(self):
        return self._statesranges

    @statesranges.setter
    def statesranges(self, value):
        self._statesranges = value

    @property
    def next_state(self):
        return self._nextstates

    @next_state.setter
    def next_state(self, val):
        self._nextstates = val

    @property
    def prev_state(self):
        return self._prevstates

    @prev_state.setter
    def prev_state(self, val):
        self._prevstates = val

    @property
    def init_state(self):
        return self._initstate

    @init_state.setter
    def init_state(self, val):
        self._initstate = val

    @property
    def actions(self):
        return self._actions

    @actions.setter
    def actions(self, val):
        self._actions = val

    @property
    def actionsranges(self):
        return self._actionsranges

    @actionsranges.setter
    def actionsranges(self, value):
        self._actionsranges = value

    @property
    def derived(self):
        return self._derived

    @derived.setter
    def derived(self, val):
        self._derived = val

    @property
    def interm(self):
        return self._interm

    @interm.setter
    def interm(self, val):
        self._interm = val

    @property
    def observ(self):
        return self._observ

    @observ.setter
    def observ(self, value):
        self._observ = value

    @property
    def observranges(self):
        return self._observranges

    @observranges.setter
    def observranges(self, value):
        self._observranges = value

    @property
    def cpfs(self):
        return self._cpfs

    @cpfs.setter
    def cpfs(self, val):
        self._cpfs = val

    @property
    def cpforder(self):
        return self._cpforder

    @cpforder.setter
    def cpforder(self, val):
        self._cpforder = val

    @property
    def gvar_to_cpforder(self):
        return self._gvar_to_cpforder

    @gvar_to_cpforder.setter
    def gvar_to_cpforder(self, val):
        self._gvar_to_cpforder = val
    
    @property
    def reward(self):
        return self._reward

    @reward.setter
    def reward(self, val):
        self._reward = val

    @property
    def terminals(self):
        return self._terminals

    @terminals.setter
    def terminals(self, value):
        self._terminals = value

    @property
    def preconditions(self):
        return self._preconditions

    @preconditions.setter
    def preconditions(self, val):
        self._preconditions = val

    @property
    def invariants(self):
        return self._invariants

    @invariants.setter
    def invariants(self, val):
        self._invariants = val
    
    @property
    def gvar_to_type(self):
        return self._gvar_to_type

    @gvar_to_type.setter
    def gvar_to_type(self, val):
        self._gvar_to_type = val
    
    @property
    def pvar_to_type(self):
        return self._pvar_to_type

    @pvar_to_type.setter
    def pvar_to_type(self, val):
        self._pvar_to_type = val

    @property
    def gvar_to_pvar(self):
        return self._gvar_to_pvar

    @gvar_to_pvar.setter
    def gvar_to_pvar(self, val):
        self._gvar_to_pvar = val

    @property
    def discount(self):
        return self._discount

    @discount.setter
    def discount(self, val):
        self._discount = val

    @property
    def horizon(self):
        return self._horizon

    @horizon.setter
    def horizon(self, val):
        self._horizon = val

    @property
    def max_allowed_actions(self):
        return self._max_allowed_actions

    @max_allowed_actions.setter
    def max_allowed_actions(self, val):
        self._max_allowed_actions = val
    
    @property
    def param_types(self):
        return self._param_types

    @param_types.setter
    def param_types(self, val):
        self._param_types = val
        
    @property
    def variable_types(self):
        return self._variable_types

    @variable_types.setter
    def variable_types(self, val):
        self._variable_types = val
    
    @property
    def variable_ranges(self):
        return self._variable_ranges

    @variable_ranges.setter
    def variable_ranges(self, val):
        self._variable_ranges = val
        
    @property
    def grounded_names(self):
        return self._grounded_names

    @grounded_names.setter
    def grounded_names(self, val):
        self._grounded_names = val
        
    @property
    def index_of_object(self):
        return self._index_of_object

    @index_of_object.setter
    def index_of_object(self, val):
        self._index_of_object = val
    
    @property
    def default_values(self):
        return self._default_values
    
    @default_values.setter
    def default_values(self, val):
        self._default_values = val
        
    # ===========================================================================
    # utility methods
    # ===========================================================================
    
    def ground_name(self, name: str, objects: Iterable[str]) -> str:
        '''Given a variable name and list of objects as arguments, produces a 
        grounded representation <variable>___<obj1>__<obj2>__...'''
        PRIME = PlanningModel.NEXT_STATE_SYM
        is_primed = name.endswith(PRIME)
        var = name
        if is_primed:
            var = var[:-len(PRIME)]
        if objects is not None and objects:
            objects = PlanningModel.OBJECT_SEP.join(objects)
            var += PlanningModel.FLUENT_SEP + objects
        if is_primed:
            var += PRIME
        return var
    
    def parse(self, expr: str) -> Tuple[str, List[str]]:
        '''Parses an expression of the form <name> or <name>___<type1>__<type2>...)
        into a tuple of <name>, [<type1>, <type2>, ...].'''
        PRIME = PlanningModel.NEXT_STATE_SYM
        is_primed = expr.endswith(PRIME)
        if is_primed:
            expr = expr[:-len(PRIME)]
        var, *objects = expr.split(PlanningModel.FLUENT_SEP)
        if objects:
            if len(objects) != 1:
                raise RDDLInvalidObjectError(f'Invalid pvariable expression {expr}.')
            objects = objects[0].split(PlanningModel.OBJECT_SEP)
        if is_primed:
            var += PRIME
        return var, objects
    
    def variations(self, ptypes: Iterable[str]) -> Iterable[Tuple[str, ...]]:
        '''Given a list of types, computes the Cartesian product of all object
        enumerations that corresponds to those types.'''
        if ptypes is None or not ptypes:
            return [()]
        objects_by_type = []
        for ptype in ptypes:
            objects = self.objects.get(ptype, None)
            if objects is None:
                raise RDDLTypeError(
                    f'Type <{ptype}> is not valid, '
                    f'must be one of {set(self.objects.keys())}.')
            objects_by_type.append(objects)
        return itertools.product(*objects_by_type)
    
    def ground_names(self, name: str, ptypes: Iterable[str]) -> Iterable[str]:
        '''Given a variable name and list of types, produces a new iterator
        whose elements are the grounded representations in the cartesian product 
        of all object enumerations that corresponds to those types.'''
        for objects in self.variations(ptypes):
            yield self.ground_name(name, objects)
    
    def is_free_variable(self, name: str) -> bool:
        '''Determines whether the quantity is a free variable (e.g., ?x).'''
        return name[0] == '?'
        
    def object_name(self, name: str) -> str:
        '''Extracts the internal object name representation. 
        Used to read objects with string quantification, e.g. @obj.'''
        if name[0] == '@':
            name = name[1:]
        return name
        
    def is_object(self, name: str, msg='') -> bool:
        '''Determines whether the given string points to an object.'''
        
        # this must be an object
        if name[0] == '@':
            name = name[1:]
            if name not in self.objects_rev:
                raise RDDLInvalidObjectError(
                    f'Object <@{name}> is not defined. {msg}')
            return True
        
        # this could either be an object or variable
        # object if a variable does not exist with the same name
        # object if a variable has the same name but free parameters
        # ambiguous if a variable has the same name but no free parameters
        if name in self.objects_rev:
            params = self.param_types.get(name, None)
            if params is not None and not params:
                raise RDDLInvalidObjectError(
                    f'Ambiguous reference to object or '
                    f'parameter-free variable with identical name <{name}>. {msg}')
            return True
        
        # this must be a variable or free parameter
        return False
    
    def is_compatible(self, var: str, objects: List[str]) -> bool:
        '''Determines whether or not the given variable can be evaluated
        for the given list of objects.'''
        ptypes = self.param_types.get(var, None)
        if ptypes is None:
            return False
        if objects is None:
            objects = []
        if len(ptypes) != len(objects):
            return False
        for (ptype, obj) in zip(ptypes, objects):
            ptype_of_obj = self.objects_rev.get(obj, None)
            if ptype_of_obj is None or ptype != ptype_of_obj:
                return False
        return True
    
    def is_non_fluent_expression(self, expr: Expression) -> bool:
        '''Determines whether or not expression is a non-fluent.'''
        if isinstance(expr, (tuple, list, set)):
            for arg in expr:
                if not self.is_non_fluent_expression(arg):
                    return False
            return True
        
        elif not isinstance(expr, Expression):
            return True
        
        etype, _ = expr.etype
        if etype == 'constant':
            return True
        
        elif etype == 'randomvar' or etype == 'randomvector':
            return False
        
        elif etype == 'pvar':
            name, pvars = expr.args
            
            # a free variable (e.g., ?x) is non-fluent
            if self.is_free_variable(name):
                return True
                
            # an object is non-fluent
            elif not pvars and self.is_object(name):
                return True
                        
            # variable must be well-defined and non-fluent
            else:
                
                # check well-defined
                var_type = self.variable_types.get(name, None)     
                if var_type is None:
                    raise RDDLUndefinedVariableError(
                        f'Variable <{name}> is not defined.')
                
                # check nested fluents
                if pvars is None:
                    pvars = []
                return var_type == 'non-fluent' \
                    and self.is_non_fluent_expression(pvars)
        
        else:
            return self.is_non_fluent_expression(expr.args)
    
    def indices(self, objects: Iterable[str], msg: str='') -> Tuple[int, ...]:
        '''Returns the canonical indices of a sequence of objects according to the
        order they are listed in the instance
        
        :param objects: object instances corresponding to valid types defined
        in the RDDL domain
        :param msg: an error message to print in case the conversion fails.
        '''
        index_of_obj = self.index_of_object
        try:
            return tuple(index_of_obj[obj] for obj in objects)
        except:
            for obj in objects:
                if obj not in index_of_obj:
                    raise RDDLInvalidObjectError(
                        f'Object <{obj}> is not valid, '
                        f'must be one of {set(index_of_obj.keys())}. {msg}')
    
    def object_counts(self, types: Iterable[str], msg: str='') -> Tuple[int, ...]:
        '''Returns a tuple containing the number of objects of each type.
        
        :param types: a list of RDDL types
        :param msg: an error message to print in case the calculation fails
        '''
        objects = self.objects
        try:
            return tuple(len(objects[ptype]) for ptype in types)
        except:
            for ptype in types:
                if ptype not in objects:
                    raise RDDLTypeError(
                        f'Type <{ptype}> is not valid, '
                        f'must be one of {set(objects.keys())}. {msg}')
    
    def ground_values(self, var: str, values: Iterable[Value]) -> Iterable[Tuple[str, Value]]:
        '''Produces a sequence of pairs where the first element is the 
        grounded variables of var and the second are the corresponding values
        from values array.
        
        :param var: the pvariable as it appears in RDDL
        :param values: the values of var(?...) in C-based order      
        '''
        keys = self.grounded_names[var]
        values = np.ravel(values, order='C')
        if len(keys) != values.size:
            raise RDDLInvalidNumberOfArgumentsError(
                f'Variable <{var}> requires {len(keys)} argument(s), '
                f'got {values.size}.')
        return zip(keys, values)
    
    def ground_values_from_dict(self, dict_values: Dict[str, object]) -> Dict[str, Value]:
        '''Converts a dictionary of values such as nonfluents, states, observ
        which has entries <var>: <list of values> to a grounded <var>: <value>
        form, where each variable <var> is grounded out.
        
        :param dict_values: the value dictionary, where values are scalars
        or lists in C-based order
        '''
        grounded = {}
        for (var, values) in dict_values.items():
            grounded.update(self.ground_values(var, values))
        return grounded
    
    def ground_ranges_from_dict(self, dict_ranges: Dict[str, str]) -> Dict[str, str]:
        '''Converts a dictionary of ranges such as statesranges
        which has entries <var>: <range> to a grounded <var>: <range>
        form, where each variable <var> is grounded out.
        
        :param dict_ranges: the ranges dictionary mapping variable to range
        '''
        return {name: prange 
                for (action, prange) in dict_ranges.items()
                    for name in self.grounded_names[action]}
    
    def groundnonfluents(self):
        return self.ground_values_from_dict(self.nonfluents)
    
    def groundstates(self):
        return self.ground_values_from_dict(self.states)
    
    def groundstatesranges(self):
        return self.ground_ranges_from_dict(self.statesranges)
    
    def groundactions(self):
        return self.ground_values_from_dict(self.actions)
    
    def groundactionsranges(self):
        return self.ground_ranges_from_dict(self.actionsranges)
    
    def groundobserv(self):
        return self.ground_values_from_dict(self.observ)
    
    def groundobservranges(self):
        return self.ground_ranges_from_dict(self.observranges)
    
    def print_expr(self):
        '''Returns a dictionary containing string representations of all 
        expressions in the current RDDL.
        '''
        printed = {}
        printed['cpfs'] = {name: str(expr) 
                           for (name, (_, expr)) in self.cpfs.items()}
        printed['reward'] = str(self.reward)
        printed['invariants'] = [str(expr) for expr in self.invariants]
        printed['preconditions'] = [str(expr) for expr in self.preconditions]
        printed['terminations'] = [str(expr) for expr in self.terminals]
        return printed
    
    def dump_to_stdout(self):
        '''Dumps a pretty printed representation of the current model to stdout.
        '''
        from pprint import pprint
        pprint(vars(self))

    
class RDDLGroundedModel(PlanningModel):
    '''A class representing a RDDL domain + instance in grounded form.
    '''

    def __init__(self):
        super().__init__()

